#ifndef _ITC_H
#define _ITC_H 1

#include "wordlist.h"
#include "semaphore.h"

//==========================================================================================
// ITCMessage/ITCPipe
//
// Relatively simple FIFO buffers to be used for communication between the GUI thread and the
// backend threads.
//==========================================================================================

// defines for ui-to-collector communication

#define CLASS_CHECKGROUP	0x10000001 // [1] arg1 = groupname
#define CLASS_GETBODY		0x10000002 // [1] arg1 = intmsgid
#define CLASS_DELETE		0x10000003 // [1] arg1 = intmsgid
#define CLASS_MOVE			0x10000004 // [3] arg1 = intmsgid    arg2 = foldername
#define CLASS_OPENSERVER	0x10000005 // [0]
#define CLASS_CLOSESERVER	0x10000006 // [0]
#define CLASS_EXPIRE		0x10000007 // [0]
#define CLASS_URGENT		0x80000000
#define CLASS_DIE			0x80000008 // [0]
#define CLASS_STOP			0x80000001 // [0]

// defines for collector-to-ui communication

#define CLASS_CHECKEDGROUP	0x20000001 // [1] arg1 = groupname
#define CLASS_NEWARTICLE	0x20000002 // [2] arg1 = groupname   arg2 = intmsgid
#define CLASS_REMARTICLE	0x20000003 // [2] arg1 = groupname   arg2 = intmsgid
#define CLASS_ERROR			0x20000004 // [2] arg1 = subsystem   arg2 = error msg
#define CLASS_STATUS		0x20000005 // [1] arg1 = status msg
#define CLASS_GOTBODY		0x20000006 // [1] arg1 = msgid

// defines for scheduler events

#define SCHED_SAVEBOX		0x30000001
#define SCHED_CHECKNEW		0x30000002
#define SCHED_EXPIRE		0x30000003
#define SCHED_CLEANGARBAGE	0x30000004



#define SUB_ERROR_CRITICAL  0x20000001 // critical error message
#define SUB_ERROR_LIMIT		0x20000002 // operation cannot be performed due to limitations of the underlying protocol
#define SUB_ERROR_INFORM	0x20000003 // informative message about a non-critical error event

// global defines

#define NO_SUBCLASS			0x00000000 // placeholder

class ITCMessage
{
	public:
						 ITCMessage (unsigned long, unsigned long, WordList *);
						~ITCMessage (void);
		
		ITCMessage		*Next, *Prev;
		unsigned long	 Class;
		unsigned long	 Subclass;
		time_t			 Scheduled;
		int				 Reschedule;
		WordList		*Message;
};

class ITCPipe
{
	public:
						 ITCPipe (void);
						~ITCPipe (void);
		
		ITCMessage		*GetNextMessage (void);
		ITCMessage		*GetNextMessage (unsigned long);
		void			 PostMessage (unsigned long, unsigned long, WordList *);
		void			 PostMessage (unsigned long, unsigned long, char *);
		void			 PostMessage (unsigned long, unsigned long, char *, char *);
		void			 PostMessage (unsigned long, unsigned long, char *, char *, char *);
		int				 Queue (void);

		void			 AddPriorityClass (unsigned long);
		
		bool			 UrgentPending;
		bool			 PriorityPending;
	
	private:
		unsigned long	 PriorityClasses[32];
		int				 NumPriorityClasses;
		ASemaphore		*Lock;
		ITCMessage		*First, *Last;
		int				 Count;
};

#endif
