#include <pthread.h>
#include <signal.h>
#include "semaphore.h"
#include "helpdesk.h"
#include "encode.h"
#include "wordlist.h"
#include "mysqlclient.h"
#include "smtp.h"
#include "buffer.h"
#include "formatter.h"
#include "mailbox.h"
#include "checksum.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <FL/Fl_Pixmap.H>
#include <FL/Fl_Multi_Browser.H>
#include <FL/x.H>
#include <FL/fl_ask.H>
#include <string.h>
#include "var.h"
#include "reg.h"
#include "substr.h"
#include "strdupp.h"
#include "chop.h"

#define locs(uk,nam) _USELOCALE ? _CLOCALE->Get_char(nam) : uk

int				  xwidths[6] = {45,320,175,115,0};
int				  hdrwidths[3] = {32,100,0};

class				 MailBox *MailBoxList[16];
Fl_Scroll 			*current_boxlist;
//void 				*dbhand;
char				 currentfolder[64];
MailBox 			*currentmailbox = NULL;

char				 currentfrom[256];
char				 currentsubject[256];
char				 currentreplyto[256];
char				 currentmsgid[256];
WordList			 currentmessage;
int					 currentbodystart=0;
int					 currentsigstart = 0;
int					 unreadonly = 0;

int					 folderlist_fullsize = 0;

Var					 PREFERENCES;

extern Fl_Pixmap	 pixmap_folder;
extern Fl_Pixmap	 pixmap_folder_open;
extern Fl_Pixmap	 pixmap_mbox;
extern Fl_Pixmap	 pixmap_compose;

int 				 _USELOCALE;
char 				*_LANGUAGE;
Var 				*_CLOCALE;

void cb_flipfolder (Fl_Widget *, void *);
void update_tree (Fl_Scroll *);
void cb_menu_untear (Fl_Widget *, void *);
void cb_folders_untear (Fl_Widget *, void *);
void cb_selectfolder (Fl_Button *, void *);
void cb_selectfolder2 (Fl_Button *, void *);
void cb_selectmessage (Fl_Widget *, void *);
void cb_sendmessage (Fl_Window *, struct Composer_Window *);
void cb_discardmessage (Fl_Window *, struct Composer_Window *);
void cb_selectaddress (Fl_Browser *, void *);
void cb_pasteaddress (Fl_Button *, void *);
void cb_viewall (Fl_Menu_ *, void *);
void cb_viewunread (Fl_Menu_ *, void *);
void displaymessage (MailMessage *);
void update_folder (void);
void new_folderitem (MailMessage *);

void set_status (char *);
void cb_compose_header_expand (Fl_Widget *, struct Composer_Window *);
void cb_compose_header_shrink (Fl_Widget *, struct Composer_Window *);
void cb_exit (Fl_Window *, void *);

extern void start_wizard (void);

void extract_recipients (WordList *, char *);

void cb_folders_undrop (Fl_Button *, void *);

Fl_Window 		*menu = (Fl_Window *) NULL;
Fl_Window 		*address = (Fl_Window *) NULL;

class FolderData
{
	public:
						 FolderData (MailMessage *);
						~FolderData (void);

		char 			*msgid;
		unsigned long	 chk_msgid;
		unsigned long	 chk_subject;
};

FolderData::FolderData (MailMessage *M)
{
	char *subj;
	msgid = strdupp (M->MsgID);
	chk_msgid = checksum (msgid);
	
	subj = M->Envelope->Get (ENVELOPE_SUBJECT);
	if (strncasecmp (subj, "Re: ", 4) == 0)
	{
		subj += 4;
	}
	chk_subject = checksum (subj);
}

FolderData::~FolderData (void)
{
	delete[] (msgid);
}

int				  adr_widths[2] = {120,0};

void cb_viewall (Fl_Menu_ *w, void *b)
{
	unreadonly = 0;
	update_folder();
}

void cb_viewunread (Fl_Menu_ *w, void *b)
{
	unreadonly = 1;
	update_folder();
}

void cb_hdrfull (Fl_Menu_ *w, void *b)
{
	PREFERENCES.Set_char_tree ("postoffice/state/main/fullheaders", "1");
}

void cb_hdrpart (Fl_Menu_ *w, void *b)
{
	PREFERENCES.Set_char_tree ("postoffice/state/main/fullheaders", "0");
}

void cb_address (Fl_Button *b, void *d)
{
	int i;
	char buf[256];
	WordList *W = PREFERENCES.Get_collection ("postoffice")->Get_wordlist ("addressbook");
	
	if (!address)
	{
		address = make_address_window();
		br_addresses->column_widths (adr_widths);
		for (i=0;i<W->Count;++i)
		{
			sprintf (buf, "%s\t%s", W->Get(i), PREFERENCES.Get_char_tree (3, "postoffice","addressbook",W->Get(i)));
			br_addresses->add (buf, NULL);
		}
		br_addresses->callback ((Fl_Callback *) cb_selectaddress, (void *) d);
	}
	bt_address_execute->callback ((Fl_Callback *) cb_pasteaddress, (void *) d);
	address->show();
}

struct Composer_Window
{
		Fl_Button 		*bt_address;
		Fl_Button			*bt_address2;
		Fl_Button			*bt_menu_expand;
		Fl_Button			*bt_resizeheader;
		
		Fl_Editor			*in_message;
		
		Fl_Input			*in_to;
		Fl_Input			*in_cc;
		Fl_Input			*in_groups;
		Fl_Input			*in_subject;
		
		Fl_Choice			*ch_personality;
		
		Fl_Window		*comp;
		Fl_Group			*gr_mini_buttons;
		Fl_Group			*gr_bigbuttons;
		Fl_Group			*gr_headers;
		Fl_Group			*gr_to;
		Fl_Group			*gr_cc;
		Fl_Group			*gr_groups;

		char				 msgid[256];
};

void cb_compose (Fl_Menu_ *m, void *d)
{
	Fl_Window *comp = make_compose_window();
	WordList *X = PREFERENCES.Get_collection ("postoffice")->Get_wordlist ("identities");
	int i;
	for (i=0;i<X->Count;++i)
	{
		compose_ch_personality->add (PREFERENCES.Get_collection ("postoffice")->Get_collection ("identities")->Get_collection (X->Get(i))->Get_char ("shortname"));
	}
	compose_ch_personality->value (0);

	comp->show();
	win_main->redraw();
	
	Composer_Window *W 	= new Composer_Window;
		W->bt_address 		= compose_bt_address;
		W->bt_address2		= compose_bt_address2;
		W->bt_menu_expand		= bt_compose_menu_expand;
		W->bt_resizeheader		= compose_bt_resizeheader;

		W->in_message 		= compose_in_message;
		W->in_to	  			= compose_in_to;
		W->in_cc				= compose_in_cc;
		W->in_groups			= compose_in_groups;
		W->in_subject			= compose_in_subject;
		W->ch_personality		= compose_ch_personality;
		
		W->comp		  		= comp;
		W->gr_mini_buttons	= gr_compose_mini_buttons;
		W->gr_bigbuttons		= gr_compose_bigbuttons;
		W->gr_headers			= gr_compose_headers;
		W->gr_to				= gr_compose_to;
		W->gr_cc				= gr_compose_cc;
		W->gr_groups			= gr_compose_groups;
		W->msgid[0]				= 0;
	
	compose_bt_address->callback ((Fl_Callback *) cb_address, (void *) W);
	compose_bt_send->callback ((Fl_Callback *) cb_sendmessage, (void *) W);
	compose_mbt_send->callback ((Fl_Callback *) cb_sendmessage, (void *) W);
	comp->callback ((Fl_Callback *) cb_discardmessage, (void *) W);

	
	
	bt_compose_menu_shrink->callback ((Fl_Callback *) cb_compose_menu_shrink, (void *) W);
	bt_compose_menu_expand->callback ((Fl_Callback *) cb_compose_menu_expand, (void *) W);
	compose_bt_resizeheader->callback ((Fl_Callback *) cb_compose_header_expand, (void *) W);
	if (strcasecmp (PREFERENCES.Get_char_tree ("postoffice/state/composer/toolbar"), "small") == 0)
		cb_compose_menu_shrink (bt_compose_menu_shrink, (void *) W);
	if (strcasecmp (PREFERENCES.Get_char_tree ("postoffice/state/composer/headers"), "full") == 0)
		cb_compose_header_expand (compose_bt_resizeheader, (void *) W);
}

void cb_selectaddress (Fl_Browser *b, void *d)
{
	char *selected = (char *) b->text (b->value());
	WordList *W = new WordList;
	W->MakeGeneric (selected, '\t');
	in_address_nickname->value (W->Get(0));
	in_address_email->value (W->Get(1));
}

void cb_pasteaddress (Fl_Button *b, void *d)
{
	Composer_Window *W = (Composer_Window *) d;
	printf ("[addressbook] callback window: %x\n", d);
	if (W) W->in_to->value (in_address_email->value());
	address->hide();
}

void cb_delete (Fl_Widget *w, void *d)
{
	FolderData *D;
	
	Fl_Multi_Browser *xbr = (Fl_Multi_Browser *) br_folderindex;
	int i;
	// CLASS_DELETE
	for (i=0;i<br_folderindex->size();++i)
	{
		if (xbr->selected(i+1))
		{
			printf ("[delete] msgid=%i select()=%i", xbr->data(i), xbr->selected(i));
			currentmailbox->ToCollector->PostMessage (CLASS_DELETE, NO_SUBCLASS, ((FolderData *)br_folderindex->data (i+1))->msgid);
			currentmailbox->RemoveMessage ( ((FolderData *)br_folderindex->data (i+1))->msgid);
			delete (FolderData *) br_folderindex->data (i+1);
			br_folderindex->remove (i+1);
			--i;
		}
	}
	currentmailbox->SaveMailBox();
}

void cb_refilter (Fl_Menu_ *w, void *d)
{
	FolderData *D;
	MailMessage *M;
	Fl_Multi_Browser *xbr = (Fl_Multi_Browser *) br_folderindex;
	int i;
	
	for (i=0;i<br_folderindex->size();++i)
	{
		if (xbr->selected(i+1))
		{
			M = currentmailbox->Find (((FolderData *)br_folderindex->data (i+1))->msgid);
			if (M)
			{
				delete (FolderData *) br_folderindex->data (i+1);
				br_folderindex->remove (i+1);
				M->Subscribers->UnTie();
				if (!M->HeaderNames->Count) M->LoadHeaders();
				M->Envelope->Set (ENVELOPE_GROUPS, "");
				currentmailbox->Deliver (M);
				--i;
			}
		}
	}
}

void cb_select_all (Fl_Menu_ *w, void *d)
{
	int i;
	Fl_Multi_Browser *xbr = (Fl_Multi_Browser *) br_folderindex;
	
	for (i=0;i<br_folderindex->size();++i)
	{
		xbr->select (i+1);
	}
	br_folderindex->redraw();
}

void cb_reply (Fl_Widget *w, void *d)
{
	int i;
	char buf[256];
	WordList *quoted_lines;
	AsyncBuffer *quoted_body;
	
	Fl_Window *comp = make_compose_window();
	WordList *X = PREFERENCES.Get_collection ("postoffice")->Get_wordlist ("identities");
	for (i=0;i<X->Count;++i)
	{
		compose_ch_personality->add (PREFERENCES.Get_collection ("postoffice")->Get_collection ("identities")->Get_collection (X->Get(i))->Get_char ("shortname"));
	}
	compose_ch_personality->value (0);
	win_main->redraw();
	
	comp->show();
	
	Composer_Window 	*W		= new Composer_Window;
		W->bt_address 		= compose_bt_address;
		W->bt_address2		= compose_bt_address2;
		W->bt_menu_expand		= bt_compose_menu_expand;
		W->bt_resizeheader		= compose_bt_resizeheader;

		W->in_message 		= compose_in_message;
		W->in_to	  			= compose_in_to;
		W->in_cc				= compose_in_cc;
		W->in_groups			= compose_in_groups;
		W->in_subject			= compose_in_subject;
		W->ch_personality		= compose_ch_personality;
		
		W->comp		  		= comp;
		W->gr_mini_buttons	= gr_compose_mini_buttons;
		W->gr_bigbuttons		= gr_compose_bigbuttons;
		W->gr_headers			= gr_compose_headers;
		W->gr_to				= gr_compose_to;
		W->gr_cc				= gr_compose_cc;
		W->gr_groups			= gr_compose_groups;

		strcpy (W->msgid, currentmsgid);

	
	compose_bt_address->callback ((Fl_Callback *) cb_address, (void *) W);
	compose_bt_send->callback ((Fl_Callback *) cb_sendmessage, (void *) W);
	compose_mbt_send->callback ((Fl_Callback *) cb_sendmessage, (void *) W);
	comp->callback ((Fl_Callback *) cb_discardmessage, (void *) W);
	
	bt_compose_menu_shrink->callback ((Fl_Callback *) cb_compose_menu_shrink, (void *) W);
	bt_compose_menu_expand->callback ((Fl_Callback *) cb_compose_menu_expand, (void *) W);
	compose_bt_resizeheader->callback ((Fl_Callback *) cb_compose_header_expand, (void *) W);

	if (strcasecmp (PREFERENCES.Get_char_tree ("postoffice/state/composer/toolbar"), "small") == 0)
		cb_compose_menu_shrink (bt_compose_menu_shrink, (void *) W);
	if (strcasecmp (PREFERENCES.Get_char_tree ("postoffice/state/composer/headers"), "full") == 0)
		cb_compose_header_expand (compose_bt_resizeheader, (void *) W);
	
	quoted_lines = new WordList;
	
	for (i=currentbodystart;i<currentsigstart;++i)
	{
		quoted_lines->Add (currentmessage.Get(i));
	}
	
	TextMessage *Tx = new TextMessage (quoted_lines);
	delete (quoted_lines);
	quoted_lines = Tx->Get (1);
	
	quoted_body = new AsyncBuffer (4096);
	
	quoted_body->AddF ("%s tapped some keys and produced:\n\n", currentfrom);
	
	for (i=0;i<quoted_lines->Count;++i)
	{
		quoted_body->AddF ("%s\n", quoted_lines->Get(i));
	}
	quoted_body->Add ((unsigned char) 0);
	
	delete (quoted_lines);
	
	if ( (tolower(currentsubject[0]) == 'r') && (tolower(currentsubject[1]) == 'e') && (currentsubject[2] == ':') )
		strcpy (buf, currentsubject);
	else sprintf (buf, "Re: %s", currentsubject);
	
	W->in_subject->value (buf);
	W->in_to->value (currentreplyto);
	
	W->in_message->LoadFrom ((char *) quoted_body->Data);
	delete (quoted_body);
}

void cb_discardmessage (Fl_Window *b, Composer_Window *W)
{
	W->comp->hide();
	delete (W);
}

void cb_sendmessage (Fl_Window *win, Composer_Window *W)
{
	int i;
	char *to;
	WordList *HdrNames = new WordList;
	WordList *HdrValues = new WordList;
	WordList *RCPT = new WordList;
	char emailto[256];
	char nameto[256];
	char buf[256];
	char *crsr;
	char *error;
	
	int crsrpos = 0;
	char *tmp;
	
	to = (char *) W->in_to->value();
	
	char *body = new char[W->in_message->CharacterCount()+256];
	W->in_message->SaveTo (body);

	if (strlen (W->msgid) > 0)
	{
		HdrNames->Add ("References");
		HdrValues->Add (W->msgid);
	}
	
	crsr = body;
	while ((crsr) && (*crsr))
	{
		if (*crsr == 13)
			*crsr = '\n';
		++crsr;
	}
	
	W->comp->deactivate();
	
	char *pers_from = NULL;
	char *pers_email = NULL;
	char *pers_replyto = NULL;
	
	char *pers_val = (char *) compose_ch_personality->text (compose_ch_personality->value());
	WordList *X = PREFERENCES.Get_collection ("postoffice")->Get_wordlist ("identities");
	for (i=0;i<X->Count;++i)
	{
		Var *VV = PREFERENCES.Get_collection ("postoffice")->Get_collection ("identities")->Get_collection (X->Get(i));
		if (strcasecmp (VV->Get_char ("shortname"), pers_val) == 0)
		{
			pers_from = VV->Get_char ("realname");
			pers_email= VV->Get_char ("mailfrom");
			pers_replyto = VV->Get_char ("replyto");
		}
	}
	
	if (!pers_from) pers_from = "Localhost";
	if (!pers_email) pers_email = "root@localhost";
	
	sprintf (buf, "%s <%s>", pers_from, pers_email);
	HdrNames->Add ((char *) "From");
	HdrValues->Add (buf);
	HdrNames->Add ((char *) "To");
	HdrValues->Add (W->in_to->value());
	if (strlen (W->in_cc->value()) > 0)
	{
		HdrNames->Add ((char *) "CC");
		HdrValues->Add (W->in_cc->value());
	}
	HdrNames->Add ((char *) "Subject");
	HdrValues->Add (W->in_subject->value());
	if (strlen (pers_replyto) > 0)
	{
		HdrNames->Add ((char *) "Reply-To");
		HdrValues->Add (pers_replyto);
	}
	printf ("to: %s\n", W->in_to->value());
	printf ("cc: %s\n", W->in_cc->value());
	extract_recipients (RCPT, W->in_to->value());
	extract_recipients (RCPT, W->in_cc->value());
	
	error = send_message_smtp   (HdrNames,
							HdrValues,
							pers_email,
							RCPT,
							(char *) body);
	
	if (error) fl_alert (error);
	else
	{
		W->comp->hide();
		delete (W->comp);
		delete (W);
	}
	delete (HdrNames);
	delete (HdrValues);
	delete (RCPT);
}

void cb_menu_tearoff (Fl_Button *b, void *d)
{
	gr_buttonbar->hide();
	tile_all->resize (tile_all->x(),20,tile_all->w(),tile_all->h()+20);
	if (!menu) menu = make_menuwindow();
	menu->show();
	menu->callback ((Fl_Callback *) cb_menu_untear);
}

void cb_menu_untear (Fl_Widget *w, void *d)
{
	tile_all->resize (tile_all->x(),40,tile_all->w(), tile_all->h()-20);
	gr_buttonbar->show();
	gr_buttonbar->redraw();
	gr_buttons->redraw();
	tile_lists->redraw();
	menu->hide();
}

int owidth = 0;
int oheight = 0;
int oldx = 0;
Fl_Window *wfolders = (Fl_Window *) NULL;

void cb_folders_tearoff (Fl_Button *b, void *d)
{
	owidth = gr_folderlist->w();
	oheight = gr_folderlist->h();
	gr_folderlist->hide();
	if (!folderlist_fullsize)
	{
		oldx = gr_folderindex->x();
		gr_folderindex->position(0,gr_folderindex->y());
		gr_folderindex->size(gr_folderindex->w()+oldx,gr_folderindex->h());
		gr_folderlist->size (0,oheight);
		tile_lists->redraw();
	}
	else
	{
		oldx = tile_all->x();
		((Fl_Widget *)tile_all)->position(0,tile_all->y());
		tile_all->size (tile_all->w()+oldx, tile_all->h());
		gr_folderlist->size (0,oheight);
		tile_lists->redraw();
	}
	wfolders->show();
	wfolders->callback ((Fl_Callback *) cb_folders_untear);
	current_boxlist = br_winfolderlist;
	update_tree (current_boxlist);
}

void cb_folders_untear (Fl_Widget *w, void *d)
{
	if (!folderlist_fullsize)
	{
		gr_folderlist->size (owidth,gr_folderindex->h());
		gr_folderindex->position(oldx,gr_folderindex->y());
		gr_folderindex->size(gr_folderindex->w()-oldx,gr_folderindex->h());
	}
	else
	{
		gr_folderlist->size (owidth,tile_all->h());
		((Fl_Widget *)tile_all)->position (oldx, tile_all->y());
		tile_all->size (tile_all->w()-oldx, tile_all->h());
	}
	gr_folderlist->show();
	tile_lists->redraw();
	tile_all->redraw();
	wfolders->hide();
	current_boxlist = br_folderlist;
	update_tree (current_boxlist);
}

int lwidth = 0;

void cb_updatewidth (Fl_Tile *t, void *d)
{
}

void ti_widthcheck (void *x)
{
	xwidths[1] = gr_folderindex->w()-665+320;
	if (xwidths[1] != lwidth)
	{
		br_folderindex->redraw();
		lwidth = xwidths[1];
		Fl::add_timeout(0.02,ti_widthcheck);
	}
	else
	{
		Fl::add_timeout(0.5,ti_widthcheck);
	}
}

static int my_handler (int event)
{
	cb_updatewidth ((Fl_Tile *) NULL, (void *) NULL);
	return 0;
}

void cb_flipfolder (Fl_Widget *w, void *d)
{
	long idx = (long) d;
	int state = PREFERENCES.Get_collection ("postoffice")->Get_collection ("mailboxes")->Get_collection (MailBoxList[idx]->Name)->Get_int ("open");
	printf ("state: postoffice/mailboxes/%s/open = %i\n", MailBoxList[idx]->Name, state);
	if (state) PREFERENCES.Get_collection ("postoffice")->Get_collection ("mailboxes")->Get_collection (MailBoxList[idx]->Name)->Set_int ("open", 0); 
	else PREFERENCES.Get_collection ("postoffice")->Get_collection ("mailboxes")->Get_collection (MailBoxList[idx]->Name)->Set_int ("open", 1); 
	//state = 1-state;
	MailBoxList[idx]->Open = (state);
	update_tree(current_boxlist);
}

void update_tree (Fl_Scroll *scr)
{
	int i;
	int x,y;
	int isselected = 0;
	Fl_Button *bt;

	i = scr->children()-1;
	while (i>=2)
	{
		scr->remove(scr->child(i-2));
		i--;
	}

	y = scr->y()+5;
	x = 0;

	scr->begin();
	Fl_Button *b = new Fl_Button(2,scr->y()+2,3,3);
	b->box(FL_NO_BOX);
	
	while (MailBoxList[x])
	{
		isselected = 0;
		if (currentmailbox)
		{
			if (currentmailbox == MailBoxList[x]) isselected = 1;
		}
		MailBoxList[x]->Open = PREFERENCES.Get_collection ("postoffice")->Get_collection ("mailboxes")->Get_collection (MailBoxList[x]->Name)->Get_int ("open");
		{
			Fl_Button *b = bt = new Fl_Button(5,y,20,15);
			b->box(FL_NO_BOX);
			if (MailBoxList[x]->Open) pixmap_folder_open.label(b);
			else pixmap_folder.label(b);
			b->argument(x);
			b->callback (cb_flipfolder);
		}
		{
			Fl_Button *b = new Fl_Button(25,y+1,150,15,MailBoxList[x]->Name);
			if (isselected) b->labelfont(FL_HELVETICA_BOLD);
			else b->labelfont(FL_HELVETICA);
			b->labelsize(11);
			b->box(FL_NO_BOX);
			b->align(FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
			b->argument(x);
			b->callback (cb_flipfolder);
		}
		y += 15;
		if (MailBoxList[x]->Open)
		{
			MailView *V = MailBoxList[x]->VFirst;
			while (V)
			{
				{
					Fl_Button *b = bt = new Fl_Button(50,y+1,150,15,V->Name);
					b->labelsize(11);
					if (isselected && !strcasecmp (currentfolder, V->Name)) b->labelfont(1);
					b->color(55,55);
					b->box(FL_NO_BOX);
					b->down_box(FL_NO_BOX);
					b->align(FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
					b->user_data ( (void *) MailBoxList[x] );
					b->callback ((Fl_Callback *) cb_selectfolder);
				}
				{
					Fl_Button *b = new Fl_Button(30,y,15,15);
					b->box(FL_NO_BOX);
					b->color (55,55);
					b->down_box(FL_NO_BOX);
					b->user_data ( (void *) bt );
					b->callback ((Fl_Callback *) cb_selectfolder2);
					pixmap_mbox.label(b);
					y += 15;
				}
				V = V->Next;
			}
		}
		++x;
	}
	scr->end();
	scr->redraw();
	Fl::flush();
}

ITCPipe *PIPE;

int __ARGC;
char **__ARGV;

int main (int argc, char *argv[])
{
	__ARGC = argc;
	__ARGV = argv;
	
	char buf[512];
	Fl_Browser *o;
	Fl_Window *W, *spl;
	int i,j;
	PIPE = new ITCPipe;
	MailView *Vw;
	MailMessage *Ms;
	struct stat mystat;

	signal (SIGPIPE, SIG_IGN);
	
	spl = make_splash_screen();
	spl->position ((Fl::w()-spl->w())/2,(Fl::h()-spl->h())/2);

	sprintf (buf, "%s/.postoffice", getenv ("HOME"));
	if (stat (buf, &mystat)!=0)
	{
		mkdir (buf, 0700);
		start_wizard();
	}


	spl->show();
	sl_splash->value (10);
	sl_splash->label ("   Loading preferences");
	Fl::wait(0.50); Fl::flush(); Fl::wait(0.10);

	sprintf (buf, "%s/.postoffice/resource.rc", getenv ("HOME"));
	PREFERENCES.Load (buf);
	_USELOCALE = atoi (PREFERENCES.Get_char_tree ("postoffice/international/uselocale"));
	_LANGUAGE = PREFERENCES.Get_char_tree ("postoffice/international/language");
	_CLOCALE = PREFERENCES.Get_collection ("locale")->Get_collection (_LANGUAGE)->Get_collection ("postoffice");
	sl_splash->value(20);
	sl_splash->label (locs("   Loading mailboxes","loading_mailboxes"));
	Fl::wait(0.1);  Fl::flush(); Fl::wait(0.10);
	
	W = make_window();
	Fl::add_timeout(0.1,ti_widthcheck);
	

	set_status (locs("Ready for action...","ready_for_action")); //bt_status->redraw();

	o = br_message;
   	o->column_widths (hdrwidths);
	
	Var *VBoxes = PREFERENCES.Get_collection ("postoffice")->Get_collection ("mailboxes");
	WordList *MBoxes = VBoxes->GatherNames();
	
	for (i=0;i<MBoxes->Count;++i)
	{
		sl_splash->value(30+((60*i)/MBoxes->Count));
		sprintf (buf, locs("   Opening mailbox: %s","opening_mailbox"), MBoxes->Get(i));
		sl_splash->label (buf);
		Fl::wait(0.1);  Fl::flush(); Fl::wait(0.05);
		printf ("[startup] Creating Mailbox: %s\n", MBoxes->Get(i));
		MailBoxList[i] = new MailBox (PIPE, MBoxes->Get(i));
	}
	MailBoxList[i] = (MailBox *) NULL;
	delete (MBoxes);

	sl_splash->value(100);
	sl_splash->label (locs("   Starting Program","starting_program"));
	Fl::wait(0.1);  Fl::flush(); Fl::wait(0.1);

	br_folderindex->callback ((Fl_Callback *) cb_selectmessage);

	current_boxlist = br_folderlist;
	update_tree (current_boxlist);
	
	if (!wfolders) wfolders = make_folderlist_window();

	win_splash->hide();
	
	if (strcasecmp (PREFERENCES.Get_char_tree ("postoffice/state/main/toolbar"), "small") == 0)
	{
		cb_menu_shrink ((Fl_Button *) NULL, NULL);
	}

	if (strcasecmp (PREFERENCES.Get_char_tree ("postoffice/state/main/mailboxtile"), "full") == 0)
	{
		cb_folders_dropdown ((Fl_Button *) NULL, NULL);
	}
	W->callback ( (Fl_Callback *) cb_exit, NULL);

	WordList *Gm = PREFERENCES.Get_collection ("postoffice")->Get_collection ("state")->Get_collection ("main")->Get_wordlist ("geometry");
	if (Gm->Count == 4)
	{
		W->resize (Gm->GetInt(0), Gm->GetInt(1), Gm->GetInt(2), Gm->GetInt(3));
	}
	
	W->show (argc, argv);

	o = br_folderindex;
    	o->column_widths (xwidths);

	sprintf (currentfolder, "Inbox");
	
	ITCMessage *M;
	MailMessage *MS;
	char *errorstring;

	double wait_time = 0.7;

	while (1)
	{
		Fl::wait(wait_time);
		if ((PIPE->Queue() > 2) && (wait_time > 0.02)) wait_time = 0.6 * wait_time;
		if ((PIPE->Queue() < 1) && (wait_time < 0.7)) wait_time = 1.2 * wait_time;
		if (M = PIPE->GetNextMessage())
		{
			switch (M->Class)
			{
				case CLASS_GOTBODY :
					currentmailbox->Lock->LockR();
					MS = currentmailbox->Find (M->Message->Get(0));
					if (MS) displaymessage (MS);
					currentmailbox->Lock->UnlockR();
					break;

				case CLASS_CHECKEDGROUP :
					//update_folder();
					break;

				case CLASS_NEWARTICLE:
					if (currentmailbox)
					{
						if (strcasecmp (currentfolder, M->Message->Get(0)) == 0)
						{
							Vw = currentmailbox->GetView (currentfolder);
							if (Vw)
							{
								currentmailbox->Lock->LockR();
								if (Ms = currentmailbox->Find (M->Message->Get(1)))
								{
									new_folderitem (Ms);
								}
								currentmailbox->Lock->UnlockR();
							}
						}
					}
					break;

				case CLASS_ERROR :
					errorstring = new char [strlen (M->Message->Get(0)) + strlen (M->Message->Get(1)) + 64];
					sprintf (errorstring, "Error in [%s]\n%s", M->Message->Get(0), M->Message->Get(1));
					if (M->Subclass == SUB_ERROR_CRITICAL) fl_alert ((const char *) errorstring);
					printf ("\n*** ERROR ***\n%s\n", errorstring);

					delete (errorstring);
					break;

				case CLASS_STATUS:
					set_status (M->Message->Get(0));
					break;
			}
			delete (M);
		}
	}
}

time_t lastclick = 0;

void cb_selectfolder (Fl_Button *b, void *v)
{
	if ((strcmp (currentfolder, b->label()) == 0) && (currentmailbox == (MailBox *) b->user_data()))
	{
		if ((time (NULL) - lastclick) == 0)
		{
			cb_check_mail (NULL, NULL);
			update_folder();
		}
		lastclick = time (NULL);
		
	}
	else
	{
		strcpy (currentfolder, b->label());
		currentmailbox = (MailBox *) b->user_data();
		update_tree (current_boxlist);
		update_folder();
		lastclick = time (NULL);
	}
}

void cb_selectfolder2 (Fl_Button *b, void *v)
{
	cb_selectfolder ((Fl_Button *)v, NULL);
}

void cb_check_mail (Fl_Widget *m, void *blah)
{
	if (currentmailbox) currentmailbox->ToCollector->PostMessage (CLASS_CHECKGROUP, NO_SUBCLASS, currentfolder);
	//if (currentmailbox) currentmailbox->Handler->CheckGroup ("inbox");
	return NULL;
}

void cb_expire (Fl_Menu_ *m, void *blah)
{
	if (currentmailbox) currentmailbox->ToCollector->PostMessage (CLASS_EXPIRE, NO_SUBCLASS, "*");
}

void update_folder (void)
{
	int i;
	int flip = 0;
	MailMessage *M;
	MailView *V = currentmailbox->GetView (currentfolder);

	for (i=0;i<br_folderindex->size();++i)
	{
		delete ((FolderData *) br_folderindex->data(i));
	}
	br_folderindex->clear();

	set_status ("Threading...");

	for (i=0;i<V->Count();++i)
	{
		flip = (++flip) & 0xff;
		if (!flip) Fl::wait (0.05);
		M = V->GetMessage (i);
		if (M)
		{
			if (unreadonly)
			{
				if (! (M->Flags & MMFL_HAVEBODY)) new_folderitem (M);
			}
			else new_folderitem (M);
			V->DoneWithMessage();
		}
	}
	
	set_status (locs("Done...","status_done"));
}

void new_folderitem (MailMessage *M)
{
	MailMessage 	*MM;
	char 			*oursubject, *thesubject;
	char			 buf[512];
	int				 i;
	int				 read = (M->Envelope->GetInt (ENVELOPE_I_READ)) ? 4 : 0;
	char 			*spc;
	unsigned long	 chk_oursubject;

	char *from = M->Envelope->Get (ENVELOPE_NAMEFROM);
	if (strlen (from) < 2) from = M->Envelope->Get (ENVELOPE_MAILFROM);

	if (! (M->Flags & MMFL_HAVEBODY)) read = 38;

	int position = -1;
	oursubject = M->Envelope->Get (ENVELOPE_SUBJECT);
	if (strncasecmp (oursubject, "Re: ", 4) == 0)
		chk_oursubject = checksum (oursubject+4);
	else chk_oursubject = checksum (oursubject);

	if (strlen (oursubject) > 4)
	{
		for (i=br_folderindex->size()-1;i>=0;--i)
		{
			FolderData *D = (FolderData *) br_folderindex->data(i+1);
			
			if (D->chk_subject == chk_oursubject)
			{
				MM = M->Parent->Find ( ((FolderData *) br_folderindex->data(i+1))->msgid);
				thesubject = MM->Envelope->Get (ENVELOPE_SUBJECT);

				if (strlen (thesubject) > 4)
				{
					if (strstr (thesubject, oursubject+4))
					{
						if (strcmp (thesubject, oursubject+4) == 0) // this message is reply to stored message
						{
							position = i+2;
							i = -1; //i = br_folderindex->size();
						}
						else if (strcmp (thesubject+4, oursubject) == 0) // stored message is reply to this message
						{
							position = i+1;
							i = -1; //br_folderindex->size();
						}
						else if (strcmp (thesubject, oursubject) == 0) // guess we're both replies
						{
							position = i+2;
							i = -1; //i = br_folderindex->size();
						}
					}
				}
			}
		}
	}

	if (strncasecmp (oursubject, "Re: ", 4) == 0)
	{
		spc = "   ";
	}
	else spc = "";

	sprintf (buf, "@C%i@.%d\t@C%i@.%s%s\t@C%i@.%s\t@C%i@.%s",
				  read,
				  br_folderindex->size(),
				  read,
				  spc,
				  M->Envelope->Get (ENVELOPE_SUBJECT),
				  read,
				  from,
				  read,
				  M->Envelope->Get (ENVELOPE_DATE));

				  
	if (position < 0) br_folderindex->add (buf, (void *) new FolderData (M));
	else br_folderindex->insert (position, buf, (void *) new FolderData (M));
}

void cb_selectmessage (Fl_Widget *w, void *v)
{
	char			*msgid =  ((FolderData *) br_folderindex->data (br_folderindex->value()))->msgid;
	
	
	sld_progress->value (5);
	//set_status (locs("Downloading message...","downloading_message")); Fl::wait(0.005);

	if (currentmailbox)
	{
		if (strlen (currentfolder))
		{
			currentmailbox->Lock->LockR();
			MailMessage *M = currentmailbox->Find (msgid);
			if (M)
			{
				if (M->Flags & MMFL_HAVEBODY)
				{
					//printf ("body cached\n");
					if (! (M->Flags & MMFL_INRAM))
						M->LoadBody();

					displaymessage (M);
				}
				else
				{
					currentmailbox->ToCollector->PostMessage (CLASS_GETBODY, NO_SUBCLASS, M->MsgID);
					//printf ("posted message for \"%s\"\n", M->MsgID);
				}
			}  else printf ("fail: message \"%s\" not found\n", msgid);

			currentmailbox->Lock->UnlockR();
		} else printf ("fail: no current folder\n");
	} else printf ("fail: no current mailbox\n");
}

void addheader (char *headername, char *headervalue)
{
	char  buf[180];

	if (strlen (headervalue) > 80)
	{
		char *crsr;
		char *end;
		char *lastopp;
		char *hdrline;
		int   first = 1;
		int   done = 0;
		
		crsr = end = headervalue;
		while (!done)
		{
			lastopp = crsr+79;
			while (((end-crsr)<80))
			{
				if ((*end == ' ')||(*end == ',')) lastopp = end+1;
				++end;
			}
			hdrline = substr (crsr,lastopp);
			if (first)
			{
				first = 0;
				sprintf (buf, "\t~b%s:\t%s", headername, hdrline);
				br_message->add (buf, NULL);
			}
			else
			{
				sprintf (buf, "\t\t%s", hdrline);
			}
			delete[] hdrline;
			
			if (!(*lastopp)) done = 1;
			else crsr = end = lastopp;
			if (strlen (crsr) < 80) done = 1;
		}
		sprintf (buf, "\t\t%s", crsr);
		br_message->add (buf, NULL);
	}
	else
	{
		sprintf (buf, "\t~b%s:\t%s", headername, headervalue);
		br_message->add (buf, NULL);
	}
}

void displaymessage (MailMessage *M)
{
	char			 buf[256];
	char			*body;
	char			*wrappoint;
	WordList		*dbr;
	char			*nval;
	char			*xval;
	int				 i;
	char			*ln, *xln, *oln;
	int				 inheaders = 1;
	int				 insignature = 0;
	MailPart		*P;

	sld_progress->value (20);

	currentfrom[0] = 0;
	currentsubject[0] = 0;
	currentreplyto[0] = 0;
	currentmsgid[0] = 0;
	currentmessage.Clear();

	strncpy (currentmsgid, M->MsgID, 255);
	currentmsgid[255] = 0;

	br_message->clear();
	br_message->format_char ('~');
	br_message->add (" ",NULL);

	if (M->HeaderNames->Count == 0) M->LoadHeaders();
	//printf ("flags: %08x INRAM=%08x\n", M->Flags, MMFL_INRAM);
	if (! (M->Flags & MMFL_INRAM)) M->LoadBody();

	for (i=0;i<M->HeaderNames->Count;++i)
	{
		nval = M->HeaderNames->Get (i);
		xval = M->HeaderValues->Get (i);

		if (strcasecmp (nval, "From") == 0)
		{
			addheader (locs("From","From"), xval);
			strncpy (currentfrom, xval, 255);
			currentfrom[255] = 0;
		}
		else if (strcasecmp (nval, "To") == 0)
		{
			addheader (locs("To","To"), xval);
		}
		else if (strcasecmp (nval, "Cc") == 0)
		{
			addheader (locs("Cc","Cc"), xval);
		}
		else if (strcasecmp (nval, "Date") == 0)
		{
			addheader (locs("Date","Date"), xval);
		}
		else if (strcasecmp (nval, "Subject") == 0)
		{
			addheader (locs("Subject","Subject"), xval);
			strncpy (currentsubject, xval, 255);
			currentsubject[255] = 0;
		}
		else if (strcasecmp (nval, "Newsgroups") == 0)
		{
			addheader (locs("Newsgroups","Newsgroups"), xval);
		}
		else if (strcasecmp (nval, "Organization") == 0)
		{
			addheader (locs("Organization","Organization"), xval);
		}
		else if (strcasecmp (nval, "Reply-To") == 0)
		{
			addheader (locs("Reply-To","Reply-To"), xval);
			strncpy (currentreplyto, xval, 255);
			currentreplyto[255] = 0;
		}
		else if (strcasecmp (nval, "Followup-To") == 0)
		{
			addheader (locs("Followup-To","Followup-To"), xval);
			strncpy (currentreplyto, xval, 255);
			currentreplyto[255] = 0;
		}
		else if (strcasecmp (nval, "BCC") == 0)
		{
			addheader (locs("BCC","BCC"), xval);
		}
		else if (atoi (PREFERENCES.Get_char_tree ("postoffice/state/main/fullheaders")))
		{
			addheader (nval, xval);
		}
	}

	br_message->add (" ", NULL);

	if (strlen (currentreplyto) == 0) strcpy (currentreplyto, currentfrom);

	P = M->GetPart (0);
	if (P)
	{
		body = P->Contents;
		//printf ("##################################\nPartContents:\n%s\n", body);
		sld_progress->value (35);
		currentmessage.Clear();
		currentmessage.MakeGeneric (body, '\n');
		
		sld_progress->value (50);
		for (i=0;i<currentmessage.Count-1;++i)
		{
			xln = currentmessage.Get(i);
			oln = chopwincrap (xln);
			ln = oln;
			while (ln)
			{
				int wrapped = 0;
				//printf ("[postwrap] strlen(ln)=%i", strlen (ln));
				if (strlen (ln) > 100)
				{
					wrapped = 1;
					wrappoint = strchr (ln+60, ' ');
					if (!wrappoint) wrappoint = ln+79;
					//printf ("[postwrap] wrappoint %d\n", wrappoint-ln);
					*wrappoint = 0;
				}
				char *tab;

				if ((!insignature) && (ln[0] != '>'))
				{
					if (strcmp (ln, "--") == 0) insignature=1;
					else if (strcmp (ln, "---") == 0) insignature=1;
					else if (strcmp (ln, "-- ") == 0) insignature=1;
					else if (strstr (ln, "----------")) insignature=1;
					else if (strstr (ln, "_____")) insignature=1;
					else if (strstr (ln, "_ _ _")) insignature=1;
					else if (strstr (ln, "====")) insignature=1;
					else if (strstr (ln, "-=-=")) insignature=1;
					if ((insignature) && ((currentmessage.Count - i) > 18)) insignature = 0;
				}
				if (!insignature) currentsigstart = i+1;
				if (!insignature)
				{
					if (ln[0] != '>') strcpy (buf, "\t~t~.");
					else strcpy (buf, "\t~t~C97~i~.");
				}
				else strcpy (buf, "\t~t~C41~.");
			
				strncat (buf, ln, 255);
				buf[255] = 0;
				while (tab = strchr (buf+3, '\t')) (tab[0]) = ' ';
				if (ln[0] == 0) strcat (buf, " ");
				br_message->add (buf, NULL);
				if (wrapped)
				{
					ln = wrappoint+1;
					*wrappoint = ' ';
				}
				else ln = (char *) NULL;
			}
			if (oln) delete[] (oln);
		}
	}

	if (M->Count > 1)
	{
		br_message->add (" ", NULL);
		for (i=1;i<M->Count;++i)
		{
			char *name;
			P = M->GetPart (i);
			if (P)
			{
				name = P->FileName;
				if ((!name)||(!strlen(name))) name = (char *) "Untitled";
				sprintf (buf, "\t~b%s %i:\t\"%s\" %s (%d bytes): %s", locs("Attachment","attachment"), i, name, P->ContentType, P->ContentSize, P->Description);
				br_message->add (buf, NULL);
			}
		}
	}
	else if (M->Count == 0) (M->Flags &= (!MMFL_HAVEBODY));
	
	//delete (X);
	//set_status (locs("Marking message as read...","marking_read")); Fl::wait(0.005);
	//sprintf (buf, "UPDATE messages SET msgread=1 WHERE msgid=%i", id);
	//dbr = dbdriver_execute (dbhand,buf);
	//delete (dbr);
	set_status (locs("Done...","status_done"));
}

void cb_folders_dropdown (Fl_Button *b, void *v)
{
	int mb_x, mb_y, mb_w, mb_h;
	int ml_x, ml_y, ml_w, ml_h;
	int ms_x, ms_y, ms_w, ms_h;

	PREFERENCES.Set_char_tree ("postoffice/state/main/mailboxtile","full");

	ms_x = br_message->x();
	ms_y = br_message->y();
	ms_w = br_message->w();
	ms_h = br_message->h();
	
	ml_x = gr_folderindex->x();
	ml_y = gr_folderindex->y();
	ml_h = gr_folderindex->h();
	ml_w = gr_folderindex->w();
	
	mb_x = gr_folderlist->x();
	mb_y = gr_folderlist->y();
	mb_w = gr_folderlist->w();
	mb_h = gr_folderlist->h();
	
	tile_lists->remove (gr_folderlist);
	tile_lists->remove (gr_folderindex);
	tile_all->remove (br_message);
	tile_all->remove (tile_lists);
	gr_tiles->remove (tile_all);
	
	br_message->resize (ml_x, ms_y, ml_w, ms_h);
	gr_folderlist->resize (mb_x, mb_y, mb_w, mb_h + ms_h);
	
	// tile_all == verticaal
	// tile_lists = horizontaal
	
	tile_all->resize (ml_x, ml_y, ml_w, ml_h+ms_h);
	tile_lists->resize (mb_x, mb_y, ms_w, ms_h + ml_h);
	tile_all->add (br_message);
	tile_all->add (gr_folderindex);
	tile_lists->add (gr_folderlist);
	tile_lists->add (tile_all);
	gr_tiles->add (tile_lists);
	gr_tiles->redraw();
	bt_dropfolders->label ("@-22<");
	bt_dropfolders->callback ((Fl_Callback *) cb_folders_undrop, (void *) NULL);
	folderlist_fullsize = 1;
};

void cb_folders_undrop (Fl_Button *b, void *v)
{
	int mb_x, mb_y, mb_w, mb_h;
	int ml_x, ml_y, ml_w, ml_h;
	int ms_x, ms_y, ms_w, ms_h;

	PREFERENCES.Set_char_tree ("postoffice/state/main/mailboxtile","compact");

	ms_x = br_message->x();
	ms_y = br_message->y();
	ms_w = br_message->w();
	ms_h = br_message->h();
	
	ml_x = gr_folderindex->x();
	ml_y = gr_folderindex->y();
	ml_h = gr_folderindex->h();
	ml_w = gr_folderindex->w();
	
	mb_x = gr_folderlist->x();
	mb_y = gr_folderlist->y();
	mb_w = gr_folderlist->w();
	mb_h = gr_folderlist->h();
	
	tile_all->remove (br_message);
	tile_all->remove (gr_folderindex);
	tile_lists->remove (tile_all);
	tile_lists->remove (gr_folderlist);
	gr_tiles->remove (tile_lists);
	
	tile_all->resize (mb_x, mb_y, mb_w+ms_w, mb_h);
	tile_lists->resize (mb_x, mb_y, mb_w+ms_w, ml_h);
	
	gr_folderlist->resize (mb_x, mb_y, mb_w, ml_h);
	br_message->resize (mb_x, ms_y, ms_w+mb_w, ms_h);
	
	tile_lists->add (gr_folderlist);
	tile_lists->add (gr_folderindex);
	tile_all->add (tile_lists);
	tile_all->add (br_message);
	
	gr_tiles->add (tile_all);
	
	gr_tiles->redraw();
	bt_dropfolders->label ("@-22>");
	bt_dropfolders->callback ((Fl_Callback *) cb_folders_dropdown, (void *) NULL);
	folderlist_fullsize = 0;
}

void cb_menu_shrink (Fl_Button *b, void *v)
{
	PREFERENCES.Set_char_tree ("postoffice/state/main/toolbar","small");

	gr_bigbuttons->hide();
	gr_minibuttons->show();
	gr_stats->hide();
	gr_tiles->size (gr_tiles->w(), gr_tiles->h()+gr_stats->h());
}

void cb_menu_expand (Fl_Button *b, void *v)
{
	PREFERENCES.Set_char_tree ("postoffice/state/main/toolbar","big");

	gr_minibuttons->hide();
	gr_bigbuttons->show();
	gr_tiles->size (gr_tiles->w(), gr_tiles->h()-gr_stats->h());
	gr_stats->position (0, gr_tiles->y()+gr_tiles->h());
	gr_stats->show();
}

char STATUS[512];

void set_status (char *stat)
{
	strcpy (STATUS, stat);
	out_ministatus->value (STATUS);
	bt_status->label (STATUS);
	bt_status->redraw();
}

void cb_compose_menu_shrink (Fl_Button *b, void *v)
{
	Composer_Window *W = (Composer_Window *) v;

	W->gr_bigbuttons->hide();
	W->gr_mini_buttons->show();
	W->in_message->ShowPanel (0);
	PREFERENCES.Set_char_tree ("postoffice/state/composer/toolbar","small");
}

void cb_compose_menu_expand (Fl_Button *b, void *v)
{
	Composer_Window *W = (Composer_Window *) v;
	
	W->gr_mini_buttons->hide();
	W->gr_bigbuttons->show();
	W->in_message->ShowPanel (1);
	PREFERENCES.Set_char_tree ("postoffice/state/composer/toolbar","big");
}

void cb_compose_header_expand (Fl_Widget *x, Composer_Window *W)
{
	int base_x, base_y, base_inc, top_inc;
	
	PREFERENCES.Set_char_tree ("postoffice/state/composer/headers","full");

	base_x = W->in_to->x();
	base_y = W->in_to->y();
	base_inc = W->in_subject->y() - base_y;
	top_inc = W->in_to->y() - W->gr_headers->y();
	
	W->ch_personality->show();
	W->gr_cc->show();
	W->gr_groups->show();
	
	W->gr_headers->size (W->gr_headers->w(), top_inc + 5 * base_inc);
	W->ch_personality->resize (base_x, base_y, W->ch_personality->w(), 20);
	W->gr_to->resize (base_x, base_y + 1 * base_inc, W->gr_to->w(), 20);
	W->gr_cc->resize (base_x, base_y + 2 * base_inc, W->gr_cc->w(), 20);
	W->gr_groups->resize (base_x, base_y + 3 * base_inc, W->gr_groups->w(), 20);
	W->in_subject->resize (base_x, base_y + 4 * base_inc, W->in_subject->w(), 20);
	W->in_message->hide();
	W->in_message->resize (W->in_message->x(), W->in_message->y() + 3 * base_inc, W->in_message->w(), W->in_message->h() - 3 * base_inc);	
	W->in_message->show();
	
	W->bt_resizeheader->callback ((Fl_Callback *) cb_compose_header_shrink, (void *) W);
	W->gr_headers->redraw();
	W->in_message->redraw();
}

void cb_compose_header_shrink (Fl_Widget *x, Composer_Window *W)
{
	int base_x, base_y, base_inc, top_inc;

	PREFERENCES.Set_char_tree ("postoffice/state/composer/headers","compact");

	base_x = W->ch_personality->x();
	base_y = W->ch_personality->y();
	base_inc = W->in_to->y() - base_y;
	top_inc = W->ch_personality->y() - W->gr_headers->y();
	
	W->ch_personality->hide();
	W->gr_cc->position (base_x, base_y);
	W->gr_cc->hide();
	W->gr_groups->position (base_x, base_y);
	W->gr_groups->hide();

	W->gr_to->position (base_x, base_y);
	W->in_subject->position (base_x, base_y + 1 * base_inc);

	W->gr_headers->size (W->gr_headers->w(), top_inc + 2 * base_inc);
	W->bt_resizeheader->callback ((Fl_Callback *) cb_compose_header_expand, (void *) W);
	W->in_message->hide();
	W->in_message->resize (W->in_message->x(), W->in_message->y() - 3 * base_inc, W->in_message->w(), W->in_message->h() + 3 * base_inc);	
	W->in_message->show();
	W->gr_headers->redraw();
	W->in_message->redraw();
}

void extract_recipients (WordList *list, char *rcpt)
{
	WordList *tmp = new WordList;
	WordList *tmp2 = new WordList;
	char *t;
	int i,j;
	int crsr=0;
	int cstart = 0;
	int quote = 0;

	printf ("checking out %s\n", rcpt);
	
	while (rcpt[crsr])
	{
		if ((rcpt[crsr] == ',') && (!quote))
		{
			tmp->Add_ (t = substr(rcpt, cstart, crsr-cstart));
			printf ("added %s\n", t);
			cstart = crsr+1;
			//printf ("[rcpt] extracted: %s\n", t);
		}
		else if ((rcpt[crsr] =='\"')||(rcpt[crsr] =='\''))
		{
			if ((quote) && (quote==rcpt[crsr])) quote = 0;
			else
			{
				if (!quote) quote = rcpt[crsr];
			}
		}
		++crsr;
	}
	t = substr(rcpt, cstart, crsr-cstart);
	if (strlen (t)) tmp->Add_ (t);
	else delete[] t;

	if (tmp->Count)
	{
		for (i=0;i<tmp->Count;++i)
		{
			t = tmp->Get(i);
			printf ("considering %s\n", t);
			if (wild_match ("*<*@*>", t))
			{
				t = strchr (t, '<');
				list->Add_ (substr (t, 1, strlen(t)-2));
				printf ("added %s\n", list->Get (list->Count -1));
			}
			else
			{
				tmp2->MakeGeneric (tmp->Get(i), ' ');
				for (j=0;j<tmp2->Count;++j)
				{
					if (strchr (tmp2->Get(j), '@'))
					{
						list->Add (tmp2->Get(j));
						printf ("added %s\n", list->Get (list->Count -1));
					}
				}
			}
		}
	}
}

void cb_quit (Fl_Window *, void *);

void cb_exit (Fl_Window *w, void *v)
{
	char buf[256];
	
	sprintf (buf, "%i,%i,%i,%i", w->x(), w->y(), w->w(), w->h());
	PREFERENCES.Set_char_tree ("postoffice/state/main/geometry", buf);
	
	w->hide();
	cb_quit (w, v);
}


void cb_quit (Fl_Window *b, void *v)
{
	char buf[256];
	char tag[256];
	int i=0;

	Fl_Window *e = make_exitwindow();
	e->position ((Fl::w()-e->w())/2,(Fl::h()-e->h())/2);
	e->show();
	Fl::flush(); Fl::wait (0.1); Fl::flush(); Fl::wait (0.1);
	
	sprintf (tag, "Making safety backup of preferences");
	bt_exitline->label (tag);
	bt_exitline->redraw();
	Fl::flush(); Fl::wait (0.1); Fl::flush(); Fl::wait (0.1);

	sprintf (buf, "%s/.postoffice/resource.rc.tmp", getenv ("HOME"));
	PREFERENCES.Save (buf);

	sprintf (tag, "Saving preferences");
	bt_exitline->label (tag);
	bt_exitline->redraw();
	Fl::flush(); Fl::wait (0.1); Fl::flush(); Fl::wait (0.1);

	sprintf (buf, "%s/.postoffice/resource.rc", getenv ("HOME"));
	PREFERENCES.Save (buf);
	sprintf (buf, "%s/.postoffice/resource.rc.tmp", getenv ("HOME"));
	unlink (buf);
	
	while (MailBoxList[i])
	{
		sprintf (tag, "Saving mailbox '%s'", MailBoxList[i]->Name);
		bt_exitline->label (tag);
		bt_exitline->redraw();
		Fl::flush(); Fl::wait (0.1); Fl::flush(); Fl::wait (0.1);

		MailBoxList[i]->SaveMailBox();
		
		sprintf (tag, "Terminating thread '%s'", MailBoxList[i]->Name);
		bt_exitline->label (tag);
		bt_exitline->redraw();
		Fl::flush(); Fl::wait (0.1); Fl::flush(); Fl::wait (0.1);
		
		MailBoxList[i]->Quit();
		delete (MailBoxList[i]);
		++i;
	}

	sprintf (tag, "Sending financial data to IRS");
	bt_exitline->label (tag);
	bt_exitline->redraw();
	Fl::flush(); Fl::wait (0.1); Fl::flush(); Fl::wait (1.0);
	e->hide();

	exit(0);
}
