//================================================================================
// itc.cpp - Post Office Inter Thread Communication class implementations
//
// (C) 1999 Pim van Riezen <pi@vuurwerk.nl>
//================================================================================

#include "mailbox.h"

ITCMessage::ITCMessage (unsigned long _Class, unsigned long _Subclass, WordList *_Message)
{
	Class = _Class;
	Subclass = _Subclass;
	Scheduled = 0;
	Reschedule = 0;
	Message = _Message;

	//printf ("[ITC msg] %08x %08x\n[ITC a01] %s\n[ITC a02] %s\n[ITC a03] %s\n",
	//		Class,
	//		Subclass,
	//		Message->Get(0),
	//		Message->Get(1),
	//		Message->Get(2));
}

ITCMessage::~ITCMessage (void)
{
	delete (Message);
}

ITCPipe::ITCPipe (void)
{
	Lock = new ASemaphore;
	First = Last = NULL;
	Count = 0;
}

ITCPipe::~ITCPipe (void)
{
	ITCMessage *M, *nM;
	M = First;
	while (M)
	{
		nM = M->Next;
		delete (M);
		M = nM;
	}
	First = Last = NULL;
	Count = 0;
}

ITCMessage *ITCPipe::GetNextMessage (void)
{
	ITCMessage *M;
	Lock->LockOnce();
	M = First;
	if (M)
	{
		if (M->Next)
		{
			M->Next->Prev = NULL;
			First = M->Next;
			--Count;
		}
		else
		{
			First = Last = NULL;
			Count = 0;
		}
	}
	Lock->UnlockOnce();
	return (M);
}

ITCMessage *ITCPipe::GetNextMessage (unsigned long _Class)
{
	ITCMessage *M;
	Lock->LockOnce();
	M = First;
	while (M)
	{
		if (M->Class == _Class)
		{
			if (M->Prev)
			{
				if (M->Next)
				{
					M->Next->Prev = M->Prev;
					M->Prev->Next = M->Next;
				}
				else
				{
					M->Prev->Next = NULL;
					Last = M->Prev;
				}
			}
			else
			{
				if (M->Next)
				{
					M->Next->Prev = NULL;
					First = M->Next;
				}
				else
				{
					First = Last = NULL;
				}
			}
			--Count;
			Lock->UnlockOnce();
			return (M);
		}
		M = M->Next;
	}
	Lock->UnlockOnce();
	return (NULL);
}

void ITCPipe::PostMessage (unsigned long _Class, unsigned long _Subclass, WordList *_Message)
{
	ITCMessage *M = new ITCMessage (_Class, _Subclass, _Message);
	Lock->LockOnce();
	if (First)
	{
		M->Next = NULL;
		M->Prev = Last;
		Last->Next = M;
		Last = M;
		++Count;
	}
	else
	{
		M->Next = NULL;
		M->Prev = NULL;
		First = Last = M;
		Count = 1;
	}
	Lock->UnlockOnce();
}

void ITCPipe::PostMessage (unsigned long _Class, unsigned long _Subclass, char *msg1)
{
	WordList *W = new WordList;
	W->Add (msg1);
	PostMessage (_Class, _Subclass, W);
}

void ITCPipe::PostMessage (unsigned long _Class, unsigned long _Subclass, char *msg1, char *msg2)
{
	WordList *W = new WordList;
	W->Add (msg1);
	W->Add (msg2);
	PostMessage (_Class, _Subclass, W);
}

void ITCPipe::PostMessage (unsigned long _Class, unsigned long _Subclass, char *msg1, char *msg2, char *msg3)
{
	WordList *W = new WordList;
	W->Add (msg1);
	W->Add (msg2);
	W->Add (msg3);
	PostMessage (_Class, _Subclass, W);
}

int ITCPipe::Queue (void)
{
	return Count;
}
