#include <string.h>
#include <stdlib.h>

#include "mailbox.h"
#include "mysqlclient.h"

extern Var PREFERENCES;

//================================================================================
// MailMessage
//
// Nothing more than a placeholder for mail messages when they are being transported
// to and from the mailboxhandlers.
//================================================================================

MailMessage::MailMessage (void)
{
	Headers = new WordList;
	Body = new WordList;
}

MailMessage::~MailMessage (void)
{
	delete (Headers);
	delete (Body);
}

//================================================================================
// BoxHandler
//
// The generic handler, shouldn't do much but follow the API while doing nothing.
//================================================================================

BoxHandler::~BoxHandler (void)
{
}

void BoxHandler::DisplayIndex (Fl_Browser *b, char *f, int u)
{
}

void BoxHandler::UpdateIndex (Fl_Browser *b, char *f, int i)
{
}

WordList *BoxHandler::GetMessage (int i, char *f)
{
	return ((WordList *) NULL);
}

void BoxHandler::DeleteMessage (int i, char *f)
{
}

void BoxHandler::MoveMessage (int i, char *f, char *d)
{
}

int BoxHandler::PostMessage (MailMessage *)
{
	return 0;
}

//================================================================================
// SQLHandler
//
// This is the handler for getting mail out of a MySQL table.
//================================================================================

SQLHandler::SQLHandler (MailBox *P)
{
	Parent = P;
	MsgIds = NULL;
	
	mysqlhandle = NULL;
}

SQLHandler::~SQLHandler (void)
{
	if (mysqlhandle) dbdriver_close (mysqlhandle);
}

void SQLHandler::OpenConnection (void)
{
	mysqlhandle = dbdriver_open  (PREFERENCES.Get_char_tree (4, "postoffice", "mailboxes", Parent->Name, "host"),
							3306,
							PREFERENCES.Get_char_tree (4, "postoffice", "mailboxes", Parent->Name, "user"),
							PREFERENCES.Get_char_tree (4, "postoffice", "mailboxes", Parent->Name, "pass"),
							PREFERENCES.Get_char_tree (4, "postoffice", "mailboxes", Parent->Name, "database"));	
}

void SQLHandler::CloseConnection (void)
{
	dbdriver_close (mysqlhandle);
}

void SQLHandler::CheckGroup (char *groupname)
{
	WordList *res, *col;
	char qry[256];
	int i;
	
	if (!mysql_handle) OpenConnection ();
	if (mysql_handle)
	{
		char *tbl = PREFERENCES.Get_char_tree (4, "postoffice", "mailboxes", Parent->Name, "table");
		sprintf (qry, "SELECT msgid FROM %s ORDER BY msgid");
		res = dbdriver_execute (mysql_handle, qry);
		for (i=0;i<res->Count;++i)
		{
			printf ("[sqlhandler::checkgroup] found %s\n", res->Get(i));
			if (! (Parent->Exists (res->Get(i))) )
			{
				printf ("[sqlhandler::checkgroup] doesn't exist yet\n");
				sprintf (qry, "SELECT msgid,mailfrom,namefrom,subject,msgread,received,folder FROM %s WHERE msgid=%s", tbl, res->Get(i));
				col = dbdriver_execute (mysql_handle, qry);
				if (col->Count)
				{
					printf ("[sqlhandler::checkgroup] creating MailMessage\n");
					MailMessage *M = new MailMessage (Parent);
					M->Envelope->Set (ENVELOPE_MSGID, col->Get(0));
					M->Envelope->Set (ENVELOPE_MAILFROM, col->Get(1));
					M->Envelope->Set (ENVELOPE_NAMEFROM, col->Get(2));
					M->Envelope->Set (ENVELOPE_SUBJECT, col->Get(3));
					M->Envelope->Set (ENVELOPE_I_READ, col->Get(4));
					M->Envelope->Set (ENVELOPE_DATE, col->Get(5));
					M->Envelope->Set (ENVELOPE_GROUPS, col->Get(6));
					M->MsgID = strdupp (col->Get(0));
					Parent->Deliver (M);
				}
			}
		}
	}
}

void SQLHandler::DisplayIndex (Fl_Browser *b, char *folder, int unread)
{
	char			  qry[256];
	WordList		*res, *col;
	char			  buf[512];
	int			  i;
	int			  id, read;
	char			*tm;
	
	if (MsgIds) delete (MsgIds);
	MsgIds = new WordList;
	
	printf ("[sqlhandler] display <%s (%i)>\n", folder, unread);
	
	char *tbl = PREFERENCES.Get_char_tree (4, "postoffice", "mailboxes", Parent->Name, "table");
	
	sprintf (qry, "SELECT msgid,namefrom,subject,msgread,received FROM %s WHERE folder='%s'", tbl, folder);
	if (unread) strcat (qry, " AND msgread=0");
	strcat (qry, " ORDER BY msgid");
	
	printf ("[sqlhandler] query <%s>\n", qry);
	
	res = dbdriver_execute (mysqlhandle, qry);
	
	b->clear();
	
	for (i=0; i<res->Count; ++i)
	{
		col	= res->GetWordList (i);
		id	= atoi (col->Get(0));
		read	= atoi (col->Get(3));
		tm	= col->Get(4);
		
		MsgIds->AddInt (id);
		
		if (strlen (tm) > 5) tm += 5;
		
		if (!read) read=136;
		else read=0;
		
		sprintf (buf, "@C%i@.%d\t@C%i@.%s\t@C%i@.%s\t@C%i@.%s", read, id, read, col->Get(2), read, col->Get(1), read, tm);
		b->add (buf, (void *) id);
	}
	delete (res);
}

void SQLHandler::UpdateIndex (Fl_Browser *b, char *folder, int unread)
{
	char			  qry[256];
	WordList		*res, *col;
	char			  buf[512];
	int			  i;
	int			  id, read;
	char			*tm;

	if (!MsgIds)
	{
		DisplayIndex (b, folder, unread);
	}
	else
	{
		printf ("[sqlhandler] update display <%s (%i)>\n", folder, unread);
		
		char *tbl = PREFERENCES.Get_char_tree (4, "postoffice", "mailboxes", Parent->Name, "table");
		
		sprintf (qry, "SELECT msgid,namefrom,subject,msgread,received FROM %s WHERE folder='%s'", tbl, folder);
		if (unread) strcat (qry, " AND msgread=0");
		strcat (qry, " ORDER BY msgid");
		
		printf ("[sqlhandler] query <%s>\n", qry);
		
		res = dbdriver_execute (mysqlhandle, qry);
		if (res->Count == 0)
		{
			b->clear();
			MsgIds->Clear();
		}
		
		for (i=0;i<res->Count;++i)
		{
			if (i<MsgIds->Count)
			{
				if (res->GetInt(i) != MsgIds->GetInt(i))
				{
					if (res->GetInt(i) < MsgIds->GetInt(i))
					{
						MsgIds->Trim (i);
						while (i <= b->size()) b->remove (i);
					}
					else
					{
						while ( (i < MsgIds->Count) && (res->GetInt(i) > MsgIds->GetInt(i)) )
						{
							MsgIds->Delete (i);
							b->remove (i);
						}
					}
				}
				if (i > MsgIds->Count)
				{
					col	= res->GetWordList (i);
					id	= atoi (col->Get(0));
					read	= atoi (col->Get(3));
					tm	= col->Get(4);
					
					MsgIds->AddInt (id);
					
					if (strlen (tm) > 5) tm += 5;
					
					if (!read) read=136;
					else read=0;
					
					sprintf (buf, "@C%i@.%d\t@C%i@.%s\t@C%i@.%s\t@C%i@.%s", read, id, read, col->Get(2), read, col->Get(1), read, tm);
					b->add (buf, (void *) id);
				}
			}
		}
		delete (res);
	}
}

WordList *SQLHandler::GetMessage (int idx, char *folder)
{
	char qry[256];
	WordList *res;
	char *tbl;

	tbl = PREFERENCES.Get_char_tree (4, "postoffice", "mailboxes", Parent->Name, "table");
	
	sprintf (qry, "UPDATE %s SET msgread=1 WHERE folder='%s' AND msgid=%i", tbl, folder, idx);
	res = dbdriver_execute (mysqlhandle, qry);
	if (res) delete (res);
	
	sprintf (qry, "SELECT body FROM %s WHERE folder='%s' AND msgid=%i", tbl, folder, idx);
	res = dbdriver_execute (mysqlhandle, qry);
	
	return (res);
}

void SQLHandler::DeleteMessage (int idx, char *folder)
{
	WordList 		*res;
	char			  qry[256];;
	char 			*tbl;
	int			  i;
	
	tbl = PREFERENCES.Get_char_tree (4, "postoffice", "mailboxes", Parent->Name, "table");
	
	sprintf (qry, "UPDATE %s set folder='_trash' WHERE msgid=%i AND folder='%s'", tbl, idx, folder);
	res = dbdriver_execute (mysqlhandle, qry);
	delete (res);
}

void SQLHandler::MoveMessage (int idx, char *folder, char *newfolder)
{

}

int SQLHandler::PostMessage (MailMessage *M)
{
	return 0;
}

MailBox::MailBox (char *nm)
{
	Name = new char[strlen(nm)+1];
	strcpy (Name, nm);
	Children = new WordList (PREFERENCES.Get_collection("postoffice")->Get_collection("mailboxes")->Get_collection(nm)->Get_wordlist("folders"));
	Handler = new SQLHandler (this);
	Open = 0;
}

MailBox::~MailBox (void)
{
	delete (Name);
	delete (Children);
	delete (Handler);
}
