// Initiation rites to be performed here. Is that a scroll in your toga or are you
// just happy to see me?

POP3Handler::POP3Handler (MailBox *_Parent, Var *_Parameters)
{
	Parameters = new Var (_Paremeters);
	Parent = _Parent;
	Con = new Connection (0);
	ConnectionOpen = 0;
	HighestHandled = 0;
	TransMsgid = new Var;
}

// So long and thanks for all the fish.

POP3Handler::~POP3Handler (void)
{
	if (Con)
	{
		if (ConnectionOpen) // Hmm, last one out turns off the lights.
		{
			Con->Close();
			ConnectionOpen = 0;
		}
		delete (Con);
		Con = NULL;
	}

	delete (Parameters); // We don't need pants where we're going.
	delete (TransMsgid);
}

void POP3Handler::OpenConnection (void)
{
	char *msg;
	char *pHost = Parameters->Get_char ("hostname");
	int   pPort = Parameters->Get_int ("port");
	int	  res;
	int	  state = 0;

	// Verify parameters

	if ( (strlen (pHost) == 0) || (pPort == 0) )
	{
		Parent->ToUI->PostMessage (CLASS_ERROR, SUB_ERROR_CRITICAL,
								   "pop3handler", "Invalid Hostname or Port in preferences. You blew it.");
		return;
	}

	// Send State information - An Informed User Is A Happy User(TM). Not that I care.

	msg = new char[ strlen (pHost) + 32];
	sprintf (msg, "Contacting host %s", pHost);
	Parent->ToUI->PostMessage (CLASS_STATE, NO_SUBCLASS, msg);
	delete[] (msg);	

	// Open Connection

	Con->Open (pHost, pPort);
	
	msg = new char[64];
	while (Con->Connected == ST_CONNECTING) Con->Handle (msg, 63); // Wait till it's *really* open
	delete[] (msg);

	// Verify connection

	if (Con->Connected == ST_CONNECTED) // Send state information
	{
		Parent->ToUI->PostMessage (CLASS_STATE, NO_SUBCLASS, "Connection established");
	}
	else // Connection failed, bail out with an annoying error.
	{
		Parent->ToUI->PostMessage (CLASS_ERROR, SUB_ERROR_CRITICAL,
								   "pop3handler", "Could not connect to POP3 server, try smoke signals.");
		return;
	}

	msg = new char[255];
	msg[0] = 0;

	if (CheckReply (msg, 255)) // POP3 servers say hello first
	{
		Con->SendF (10, "USER %s\n", Parameters->Get_char ("username")); // Send username, quelle surprise.
		if (CheckReply (msg, 255))
		{
			Con->SendF (10, "PASS %s\n", Parameters->Get_char ("password")); // Send password.
			if (CheckReply (msg, 255))
			{
				ConnectionOpen = 1; // The eagle has landed.
			}
		}
	}
	if (!ConnectionOpen) // If it's not open, it must be closed.
	{
		Con->Close(); // Confirm this narrow view of life.
	}
	
	delete[] (msg); // Clean up the mess
	HighestHandled = 0; // Make a fresh start
}

void POP3Handler::CloseConnection (void)
{
	if (ConnectionOpen)
	{
		Con->Close();
		Parent->ToUI->PostMessage (CLASS_STATE, NO_SUBCLASS, "Connection closed");
		ConnectionOpen = 0;
	}
}

//====================================================================================================
// Internal handler to take care of pop3 protocol replies. In case of an error reply it automatically
// takes care of informing the UI thread.
//====================================================================================================

int POP3Handler::CheckReply (char *buf, int len)
{
	int res;
	buf[0] = 0;
	
	while ((res = Con->Handle (buf, len)) == 0);
	if (buf[0] == '+') return 1;
	else
	{
		if (buf[0] == '-') // Some kind of oddbals protocol error, let the user figure it out.
		{
			char *errorstring = new char[strlen (buf+1) + 96];
			sprintf (errorstring, "I wanted it to zig, and the POPserver zagged. \"Behold,\" he "
								  "whispered, \"%s\".", buf+1);
			Parent->ToUI->PostMessage (CLASS_ERROR, SUB_ERROR_CRITICAL, "pop3handler", errorstring);
			delete[] (errorstring);
		}
	}
	return 0;
}

//====================================================================================================
// POP3 has no methods to check for a particular group. The groupname argument is therefor lost
// in the line noise. Unfortunately POP3 numbers messages in a mailbox in quite an unsafe way,
// which means we need to keep track of a temporary translation table from MsgIds to each message
// in the mailbox. Life sucks.
//====================================================================================================

void POP3Handler::CheckGroup (char *groupname)
{
	char 		msg[64];
	int  		numberofmessages;
	AsyncBuffer	msgbuffer;
	char		buf[256];
	char		msgid[128];
	int			done = 0;
	int			i;
	int			fetchbodies = 0;
	int			reply;
	int			deletemail = 0;

	int			start;

	if (!ConnectionOpen) OpenConnection(); // Open the connection on demand

	if (ConnectionOpen) // Verify the connection is now really open
	{
		Con->SendF (10, "STAT\n"); // Determine the amount of messages waiting for us
		if (CheckReply (msg, 63))
		{
			numberofmessages = atoi (msg+4);

			sprintf (buf, "%i messages on pop3 server", numberofmessages);
			Parent->ToUI->PostMessage (CLASS_STATUS, NO_SUB_CLASS,
									   "pop3handler", buf); // Let the UI thread know what's going on
			
			if (strcasecmp (Parameters->Get_char ("fetchbodies"), "Y") == 0) // Method 1: fetch entire bodies at a time
				fetchbodies = 1;

			if (strcasecmp (Parameters->Get_char ("delete"), "Y") == 0) // Delete mail from server?
				deletemail = 1;

			if (HighestHandled) start = HighestHandled-1;
			else start = 0;
			
			for (i=start;i<numberofmessages;++i) // Queue up RETR/TOP commands (breakfast serials are served)
			{
				if (checkbodies) Con->SendF (10, "RETR %i\n", i+1);
				else Con->SendF (10, "TOP %i 0\n", i+1);
			}
			for (i=start;i<numberofmessages;++i) // Iterate through each sent RETR command
			{
				if (CheckReply (msg, 63)) // See if it generated a positive reply
				{
					sprintf (buf, "retrieving message %i", i);
					Parent->ToUI->PostMessage (CLASS_STATUS, NO_SUB_CLASS, "pop3handler", buf);
					
					msgid[0] = 0; // Clean msgid
					
					while (!done) // For each line in the message
					{
						buf[0] = 0;
						if (Con->Handle (buf, 255) < 0) // Make sure the connection is still alive
						{
							done = 1;
							Con->Close();
							ConnectionOpen = 0;
							Parent->ToUI->PostMessage (CLASS_ERROR, SUB_ERROR_CRITICAL,
													   "pop3handler",
													   "There I was, collecting mail, minding my own business, and this punkass "
													   "POP3 daemon started acting like a pain, dropping connections all over.");
						}
						else
						{
							if (buf[0]) // See if we received actual data
							{
								if ((buf[0]=='.') && (buf[1] < 32)) done = 1; // End of message reached
								else
								{
									msgbuffer.Add (buf);
									if (!msgid[0]) // Check if we already have the msgid
									{
										if (wild_match ("Message-ID:*", buf)) // See of this is the message-id
										{
											char *tmp;
											
											choplf (buf);

											// Seperate value from headername

											tmp = strchr (buf,':');
											if (*tmp) ++tmp;
											while ( (*tmp==' ') || (*tmp=='\t') ) ++tmp;

											// copy over msgid
											
											strncpy (msgid, tmp, 127);
											msgid[127] = 0;

											// store the msgid in the translation table
											
											TransMsgID->Set_int (msgid, i+1);
										}
									}
								}
							}
						}
					}
					if (ConnectionOpen) // Only interested in completely transfered messages thank you
					{
						if (strlen (msgid) < 1) // In the odd case there is no msgid we'll have to make one up
						{
							sprintf (msgid, "<POP3_NoMSGID_%08i>", i); // Of course this is inherently unstable and
																	   // should not happen, but we need _some_
																	   // id for the message. Oh heck, any message
																	   // evil enough not to have a Message-ID set by
																	   // the server deserves to haunt the user,
																	   // begging him to LART the sender AND the
																	   // author of the mailerdaemon.
						}
						if (! (Parent->Have (msgid))) // Make sure this isn't a double
						{
							sprintf (buf, "processing message %i msgid %s", i, msgid);
							Parent->ToUI->PostMessage (CLASS_STATUS, NO_SUB_CLASS, "pop3handler", buf);
						
							MailMessage *X = new MailMessage (Parent, msgid);
							
							if (checkbodies) X->ProcessMessage (msgbuffer.Data);
							else X->ProcessHeaders (msgbuffer.Data);

							if (deletemail) X->Flags |= MMFL_ORIGINAL; // If we're going to delete the mail from the
																	   // server it's essential to turn on the preservation
																	   // bits to prevent the message from getting flushed away.
							
							Parent->Deliver (X);  // And now it's safely out of our hands.
							HighestHandled = i+1; // Save session state
						}
					}
				}
				if (!ConnectionOpen) i = numberofmessages; // In case things messed up, bail out
				msgbuffer.Pos = 0; // Clear the buffer
			}
			
			if (checkbodies && deletemail) // Delete mail from server?
			{
				for (i=start; i<HighestHandled; ++i) // Queue up the DELE commands
				{
					Con->SendF (10, "DELE %i\n", i+1);
				}
				for (i=start; i<HighestHandled; ++i) // And verify confirmation
				{
					if (CheckReply (msg, 63) < 0) // Connection-related error
					{
						i = HighestHandled; // Bail out and stuff
						ConnectionOpen = 0;
						Con->Close();
					}
					
				}
			}
		}
	}
}

void POP3Handler::GetBody (char *msgid)
{
	MailMessage 	*X = Parent->Find (msgid);
	char			 msg[512];
	int				 done;
	int				 headersdone;
	int				 result;
	ASyncBuffer		 msgbuffer;

	if (X) // If the msgid could be found...
	{
		if (!X->HaveBody) // And is not pointing to an already downloaded message...
		{
			if (!ConnectionOpen) // Open the connection as needed
			{
				OpenConnection();
				if (ConnectionOpen) CheckGroup (""); // And update the translation table (Ireland awards one Kludge point).
			}
			if (ConnectionOpen) // Check to see if the connection is really open now
			{
				int popid = TransMsgID->Get_int (msgid); // Lookup in translation table
				if (popid)
				{
					Con->SendF (10, "RETR %i\n", popid); // Tell the POP3 server to tango
					if (CheckReply (msg, 511)) 			 // And wait for its stern lead
					{
						done = headersdone = 0; // Reaffirm our position in the Time Space Continuum
						while (!done)
						{
							msg[0] = 0;
							if ( (result = Con->Handle (msg, 511)) < 0 ) // Bad mojo in the connection handler
							{
								ConnectionOpen = 0; // Burn our bridges
								Con->Close();
								
								Parent->ToUI->PostMessage (CLASS_ERROR, SUB_ERROR_CRITICAL, // And whine to the user
														   "pop3handler",
														   "Server disconnected during body retrieval. That bastard.");
								done = 1;
								headersdone = 0;
							}
							if (result > 0) // Runes were scribed by the server
							{
								if (headersdone) // Are we processing real data yet?
								{
									if ((msg[0]=='.')&&(msg[1]<32)) done = 1; // Beware of the Death Dot
									else msgbuffer.Add (msg); // Or feed the force
								}
								else if (strlen (msg) == 1) // Nope, is there an end to the header misery yet?
								{
									headersdone = 1;
								}
							}
						}
						if (ConnectionOpen) // Did we survive El Tango del Muerte?
						{
							X->ProcessBody (msgbuffer.Data); // Let the message absorb the data
						}
					}
				}
			}
		}
	}
	else // No such message in our database, let's moan a bit for old time's sake.
	{
		sprintf (msg, "Could not locate Message \"%s\" in internal database. You probably did not want to hear this.", msgid);
		Parent->ToUI->PostMessage (CLASS_ERROR, SUB_ERROR_CRITICAL, "pop3handler", msg);
	}
}
