#include "mailbox.h"
#include "wordlist.h"
#include "var.h"
#include "semaphore.h"
#include "chop.h"
#include "checksum.h"

#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>

extern Var PREFERENCES;

MailBox::MailBox (ITCPipe *ui, char *name)
{
	int i;
	
	Name = new char[strlen(name)+1];
	strcpy (Name, name);

	Lock = new ASemaphore ("mailbox::lock");
	VLock = new ASemaphore ("mailbox::vlock");

	Var *PREFROOT = PREFERENCES.
					Get_collection ("postoffice")->
					Get_collection ("mailboxes")->
					Get_collection (Name);

	First = Last = NULL;
	VFirst = VLast = NULL;

	ToCollector = new ITCPipe;
	ToUI = ui;
	Schedule = new Scheduler;
	
	Schedule->Schedule (120, 120, SCHED_SAVEBOX, NO_SUBCLASS, "");
	i = PREFROOT->Get_int ("interval");
	if (i) Schedule->Schedule (5, i, SCHED_CHECKNEW, NO_SUBCLASS, "");
	
	MemoryUsage = 0;
	Count = 0;
	CacheToDisk = 0;
	Collector = new MailerThread;

	char *HOME = (char *) getenv ("HOME");
	DirectoryPrefix = new char [strlen (HOME) + strlen ("/.postoffice/.box") + strlen (Name) + 3];
	sprintf (DirectoryPrefix, "%s/.postoffice/%s.box", HOME, Name);

	if (strcasecmp (PREFROOT->Get_char ("protocol"), "pop3") == 0) // POP3 box
	{
		Handler = new POP3Handler (this, PREFROOT);
	}
	else if (strcasecmp (PREFROOT->Get_char ("protocol"), "mysql") == 0)
	{
		Handler = new MySQLHandler (this, PREFROOT);
	}
	else if (strcasecmp (PREFROOT->Get_char ("protocol"), "nntp") == 0)
	{
		Handler = new NNTPHandler (this, PREFROOT);
	}

	WordList *Views = PREFROOT->Get_wordlist ("folders");

	for (i=0;i<Views->Count;++i)
	{
		CreateMailView (Views->Get(i));
	}


	Lock = new ASemaphore ("MailBox.Lock");
	VLock = new ASemaphore ("MailBox.VLock");

	LoadMailBox();
	Collector->Run((void *) this);
}

MailBox::~MailBox (void)
{
	SaveMailBox();
	Collector->Kill();
}

/*
#define ENVELOPE_I_READ			0
#define ENVELOPE_I_PARENTID		1
#define ENVELOPE_MAILFROM		2
#define ENVELOPE_NAMEFROM		3
#define ENVELOPE_SUBJECT		4
#define ENVELOPE_DATE			5
#define ENVELOPE_NEWSGROUPS		6
#define ENVELOPE_GROUPS			6
*/

void MailBox::LoadMailBox (void)
{
	FILE *IDX;
	char *fname;
	char buf[512];
	WordList Line;
	WordList Subs;
	int i;
	MailView *V;
	int delivered;
	struct stat mystat;
	
	char path[256];
	const char *HOME = getenv ("HOME");
	
	if (!HOME) HOME = "/tmp";
	
	if (stat (HOME, &mystat)==0)
	{
		sprintf (path, "%s/.postoffice", HOME);
		if (stat (path, &mystat)!=0)
		{
			mkdir (path, 0700);
		}
		strcat (path, "/");
		strcat (path, Name);
		strcat (path, ".box");
		if (stat (path, &mystat)!=0) mkdir (path, 0700);
	}

	fname = new char [strlen (DirectoryPrefix) + strlen ("index") + 16];
	sprintf (fname, "%s/index", DirectoryPrefix);

	//printf ("opening %s\n", fname);

	if (IDX = fopen (fname, "r")) // Attempt to open the index file;
	{
		while (!feof (IDX)) // Browse through the index file
		{
			buf[0] = 0;
			fgets (buf, 511, IDX); // Read a line
			if ((strlen (buf)>1)&&(!feof (IDX)))
			{
				choplf (buf);
				
				Line.MakeDB (buf); // The 9 arguments to a message in the index
				Subs.MakeGeneric (Line.Get(6),','); // The names of all views it was a part of
				
				MailMessage *M = new MailMessage (this, Line.Get(0));
				M->Envelope->SetInt (ENVELOPE_I_READ, Line.GetInt(8));
				M->Envelope->Set (ENVELOPE_I_PARENTID, Line.Get(1));
				M->Envelope->Set (ENVELOPE_MAILFROM, Line.Get(2));
				M->Envelope->Set (ENVELOPE_NAMEFROM, Line.Get(3));
				M->Envelope->Set (ENVELOPE_SUBJECT, Line.Get(4));
				M->Envelope->Set (ENVELOPE_DATE, Line.Get(5));
				M->Envelope->Set (ENVELOPE_GROUPS, Line.Get(6));
				M->Count = Line.GetInt(9);
				M->Flags = Line.GetInt(7) & (MMFL_ORIGINAL | MMFL_HAVEBODY | MMFL_ONDISK);

				// Add the message to this queue


				// Add the message to the views

				delivered = 0;
				
				VLock->LockR();
					for (i=0;i<Subs.Count;++i)
					{
						V = GetView (Subs.Get(i)); // Find the named view
						if (!V) // If it doesn't exist
						{
							printf ("%s does not exist, using Deliver()\n", Subs.Get(i));
							M->LoadHeaders();
							UnTie (M); // Undo direct deliveries
							Deliver (M); // And fall back to filter rules
							i = Subs.Count; // Exit direct delivery cycle
							delivered = 1;
						}
						else // But of course it exists
						{
							V->Associate (M); // Direct delivery
						}
					}
				VLock->UnlockR();
				if (!delivered)
				{
					Lock->LockRW(); 
						if (Last)
						{
							M->Prev = Last;
							M->Next = NULL;
							Last->Next = M;
							Last = M;
							++Count;
						}
						else
						{
							M->Prev = M->Next = NULL;
							First = Last = M;
							Count = 1;
						}
					Lock->UnlockRW();
				}
			}
		}
	}
}

void MailBox::SaveMailBox (void)
{
	struct stat mystat;
	char path[256];
	const char *HOME = getenv ("HOME");
	
	if (!HOME) HOME = "/tmp";
	
	if (stat (HOME, &mystat)==0)
	{
		sprintf (path, "%s/.postoffice", HOME);
		if (stat (path, &mystat)!=0)
		{
			mkdir (path, 0700);
		}
		strcat (path, "/");
		strcat (path, Name);
		strcat (path, ".box");
		if (stat (path, &mystat)!=0) mkdir (path, 0700);
		strcat (path, "/index");
		FILE *OUT = fopen (path, "w");
		if (OUT)
		{
			MailMessage *M;
			ViewListNode *V;
			int i;
		
			Lock->LockR();
			VLock->LockR();

			M = First;
			while (M)
			{
				fprintf (OUT, "%s|%s|%s|%s|%s|%s|",
							  M->MsgID,
							  M->Envelope->Get (ENVELOPE_I_PARENTID),
							  M->Envelope->Get (ENVELOPE_MAILFROM),
							  M->Envelope->Get (ENVELOPE_NAMEFROM),
							  M->Envelope->Get (ENVELOPE_SUBJECT),
							  M->Envelope->Get (ENVELOPE_DATE));

				for (i=0;i<M->Subscribers->Count();++i)
				{
					if (V=M->Subscribers->Get(i))
					{
						fprintf (OUT, "%s", V->View->Name);
						if (i+1 < M->Subscribers->Count()) fputc (',', OUT);
					}
				}
				fprintf (OUT, "|%i|%i|%i|%i\n", M->Flags, M->Envelope->GetInt (ENVELOPE_I_READ), M->Count, M->Envelope->GetInt (ENVELOPE_RECEIVETIME));

				//M->SaveHeaders();

				M = M->Next;
			}

			fclose (OUT);

			Lock->UnlockR();
			VLock->UnlockR();
		}
		else
		{
			ToUI->PostMessage (CLASS_ERROR, SUB_ERROR_CRITICAL, "mailboxhandler", "Could not create index file");
			printf ("error creating index file\n");
		}
	}
}

void MailBox::UnTie (MailMessage *M)
{
	MailView *V = VFirst;

	while (V) // While the fat lady's still singing
	{
		V->Unassociate (M); // This procedure is expensive and should best be avoided, alas I'm lazy today.
		V = V->Next; // And on and on and on...
	}
}

void MailBox::Deliver (MailMessage *M)
{
	if (!(Find (M->MsgID)))
	{
		Lock->LockRW();
		if (Last)
		{
			M->Next = NULL;
			M->Prev = Last;
			Last->Next = M;
			Last = M;
		}
		else
		{
			M->Next = M->Prev = NULL;
			First = Last = M;
		}
	}
	else Lock->LockRW();
	
	MailView *V;

	int delivered = 0;

	V = VFirst;

	while ((V)&&(!delivered))
	{
		delivered = V->ProcessRules (M);
		V = V->Next;
	}

	if (!delivered)
	{
		char *fld = PREFERENCES.Get_collection ("postoffice")->Get_collection ("mailboxes")->Get_collection (Name)->Get_char ("defaultfolder");
		V = GetView (fld);
		if (!V) V = GetView ("Inbox");
		if (!V) V = VFirst;
		if (V) V->Associate (M);
	}
	Lock->UnlockRW();

}

int MailBox::Have (char *msgid)
{
	unsigned long check = checksum (msgid);
	Lock->LockR();
		MailMessage *M = First;
		while (M)
		{
			if (M->cMsgID == check)
			{
				if (strcasecmp (msgid, M->MsgID) == 0)
				{
					Lock->UnlockR();
					return 1;
				}
			}
			M = M->Next;
		}
	Lock->UnlockR();
	return 0;
}

void MailBox::CreateMailView (char *name)
{
	MailView *V = new MailView (this, name);
	VLock->LockRW();
		if (VLast)
		{
			V->Next = NULL;
			V->Prev = VLast;
			VLast->Next = V;
			VLast = V;
		}
		else
		{
			V->Next = V->Prev = NULL;
			VFirst = VLast = V;
		}
	VLock->UnlockRW();
}

MailView *MailBox::GetView (char *Name)
{
	MailView *V;

	VLock->LockR();
		V = VFirst;
		while (V)
		{
			if (strcasecmp (V->Name, Name) == 0)
			{
				VLock->UnlockR();
				return (V);
			}
			V = V->Next;
		}
	VLock->UnlockR();
	return ((MailView *) NULL);
}

MailMessage *MailBox::Find (char *msgid)
{
	unsigned long check = checksum (msgid);
	Lock->LockR();
		MailMessage *M = First;
		while (M)
		{
			if (M->cMsgID == check)
			{
				if (strcasecmp (msgid, M->MsgID) == 0)
				{
					Lock->UnlockR();
					return M;
				}
			}
			M = M->Next;
		}
	Lock->UnlockR();
	return (MailMessage *)NULL;
}

void MailBox::Quit (void)
{
	int hitcount = 0;
	
	printf ("quitting mailbox started=%i\n", Collector->Started);
	if (Collector->Started)
	{
		ToCollector->PostMessage (CLASS_DIE, NO_SUBCLASS, "");
		printf (">>> Sent Death."); fflush (stdout);
		while (Collector->Started)
		{
			printf ("."); fflush (stdout);
			if (Collector->Started) sleep(1);
			if (++hitcount > 15)
			{
				printf ("<timeout>\nKilling.. "); fflush (stdout);
				Collector->Kill();
				Collector->Started = 0;
			}
		}
		printf ("done\nThis is a dead thread.\n");
	}
}

void MailBox::RemoveMessage (char *msgid)
{
	MailMessage *M = Find (msgid);
	if (M)
	{
		Lock->LockRW();
			
			M->Subscribers->UnTie();
			
			if (M->Flags & (MMFL_HAVEBODY | MMFL_ONDISK))
			{
				M->RemoveFiles();
			}
			
			if (M->Prev)
			{
				if (M->Next)
				{
					M->Next->Prev = M->Prev;
					M->Prev->Next = M->Next;
				}
				else
				{
					M->Prev->Next = NULL;
					Last = M->Prev;
				}
			}
			else
			{
				if (M->Next)
				{
					M->Next->Prev = NULL;
					First = M->Next;
				}
				else
				{
					First = Last = NULL;
				}
			}
			--Count;
			delete (M);
		
		Lock->UnlockRW();
	}
}
