#ifndef VAR_H
#define VAR_H 1
#include <sys/types.h>
#include <stdio.h>
#include "wordlist.h"

#define VT_INT 1
#define VT_SHORT 2
#define VT_LONG 3
#define VT_CHAR 4
#define VT_BINARY 5
#define VT_WORDLIST 6
#define VT_POINTER 7
#define VT_COLLECTION 8

#define MAXVARNAMELEN 128

class Value
{
	public:
						 Value (char *);
						~Value (void);
		Value			*Prev, *Next;
		Value			*Parent, *HigherNode, *SameNode, *LowerNode;
		
		void			 Rename (char *);
		void			 CopyTo (class Var *);

		void			 Set_collection (Var *);
		void			 Set_int (int);
		void			 Set_long (u_long);
		void			 Set_char (char *);
		void			 Set_short (u_short);
		void			 Set_binary (char *, size_t);
		void			 Set_wordlist (WordList *);
		void			 Set_pointer (void *);
		int				 Get_int (void);
		u_long			 Get_long (void);
		char			*Get_char (void);
		u_short			 Get_short (void);
		char			*Get_binary (void);
		size_t			 Get_size (void);
		WordList		*Get_wordlist (void);
		void			*Get_pointer (void);
		u_short			 Get_type (void);
		class Var		*Get_collection (void);
		
		char			 Name[MAXVARNAMELEN];
		u_long			 Checksum;
		
		char			 Altered;
		
	private:
		u_short			 TYPE;
		char			*cValue;
		u_short			 sValue;
		int				 iValue;
		u_long			 lValue;
		size_t			 BSize;
		WordList		*wValue;
		void			*Pointer;
		Var				*vValue;
};

class Var
{
	public:
						 Var (void);
						 Var (Var *);
						~Var (void);
		void			 Set_int (char *, int);
		void			 Set_long (char *, u_long);
		void			 Set_char (char *, char *);
		void			 Set_short (char *, u_short);
		void			 Set_binary (char *, char *, size_t);
		void			 Set_wordlist (char *, WordList *);
		
		int				 Get_int (char *);
		u_long			 Get_long (char *);
		char			*Get_char (char *);
		u_short			 Get_short (char *);
		char			*Get_binary (char *);
		size_t			 Get_size (char *);
		u_short			 Get_type (char *);
		Var				*Get_collection (char *);
		WordList		*Get_wordlist (char *);
		
		void			 Rename (char *, char *);
		void			 Delete (char *);

		char			*Get_char_tree (char *);
		char			*Get_char_tree (char *, char);
		char			*Get_char_tree (int, ...);
		void			 Set_char_tree (char *, char *);
		void			 Set_char_tree (char *, int, char *, ...);
		char			*Get_char_backmatch (char *, char *);
		WordList		*Get_wordlist_backmatch (char *, char *);

		WordList		*GatherNames (void);
		WordList		*GatherValues (void);
		
		void			 Save (char *);
		void			 Save (FILE *, int);
		void			 Load (char *);
		void			 Load (FILE *);
		void			 CopyTo (Var *);
		void			 CopyFrom (Var *);

		int				 Count;
		Value			*First, *Last;
		Value			*Find (char *);
		void			 Clear (void);
	
	private:
		Value			*_Find (char *);

		u_long			 CacheChecksum[16];
		Value			*CacheValue[16];
		
		void			 Add (Value *);
		void			 Relink (Value *);
};



void reclaim_var (void *);

#endif
