#include "wordlist.h"
#include <stdio.h>
#include <string.h>

//===============================================================================
// CONSTRUCTOR	Word
// SYNOPSIS		Currently no initialization necessary
//===============================================================================

Word::Word (void)
{
	Contents = NULL;
	WContents = NULL;
	IntSet = 0;
	IContents = 0;
}

//===============================================================================
// DESTRUCTOR	Word
// SYNOPSIS		Currently no actions performed.
//===============================================================================

Word::~Word (void)
{
	if (Contents) delete[] (Contents);
	if (WContents) delete (WContents);
}

//===============================================================================
// CONSTRUCTOR	WordList
// SYNOPSIS		Initializes memory and variables.
//===============================================================================

WordList::WordList (void)
{
	Integrity[0]=13; Integrity[1]=37;
	First = NULL;
	Last = NULL;
	Cache = NULL;
	CachePos = -3;
	//Enumerate = NULL;
	Protect = 0;
	//Status = 0;
	Count = 0;
}

//===============================================================================
// CONSTRUCTOR	WordList
// SYNOPSIS		Initializes memory and variables. Copies another wordlist into
//				the newly created structure.
//===============================================================================

WordList::WordList (WordList *Org)
{
	Integrity[0]=13; Integrity[1]=37;
	First = NULL;
	Last = NULL;
	//Enumerate = NULL;
	Protect = 0;
	//Status = 0;
	Count = 0;
	Cache = NULL;
	CachePos = -3;
	WordList *T;
	
	int i = 0;
	
	for (i=0;i<Org->Count;++i)
	{
		if (Org->IsWordList (i)) AddWordList (Org->GetWordList (i));
		else Add (Org->Get(i));
	}
}

//===============================================================================
// DESTRUCTOR	WordList
// SYNOPSIS		Frees up memory.
//===============================================================================

WordList::~WordList (void)
{
	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		Clear ();
	}
}

//===============================================================================
// METHOD		Make
// SYNOPSIS		Turns a string into a wordlist. The default seperation scheme is
//				used (rfc1459/irc style, split by spaces up to the first colon).
//
// ARGUMENTS	theLine		(char *)	String that needs to be split up
//
// RETURNS		method returns no data.
//===============================================================================

void WordList::Make (char *theLine)
{
	char            myWord[2048];
	int             SourcePosition = 0;
	int             DestPosition = 0;

	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		myWord[0] = 0;

		Clear ();
		First = NULL;
		Last = NULL;

		if (theLine[0] == ':')
			++SourcePosition;

		while (theLine[SourcePosition] != 0)
		{
			while ((theLine[SourcePosition] != 0) && (theLine[SourcePosition] != ' ') && (DestPosition < 1023))
			{
				if ((theLine[SourcePosition] > 3) && (theLine[SourcePosition] < 30))
					++SourcePosition;
				else
					myWord[DestPosition++] = theLine[SourcePosition++];
			}
			myWord[DestPosition] = 0;
			Add (myWord);
			DestPosition = 0;
			if (theLine[SourcePosition] != 0)
				++SourcePosition;
			if (theLine[SourcePosition] == ':')
			{
				//++SourcePosition;
				while (theLine[SourcePosition] != 0)
				{
					if ((theLine[SourcePosition] > 3) && (theLine[SourcePosition] < 30))
						++SourcePosition;
					else
						myWord[DestPosition++] = theLine[SourcePosition++];
				}
				myWord[DestPosition] = 0;
				Add (myWord);
			}
		}
	}
}

//===============================================================================
// METHOD		MakeDB
// SYNOPSIS		Turns a string into a wordlist. The database seperation scheme is
//				used (seperated by '|', backslash is escape character).
//
// ARGUMENTS	theLine		(char *)	String that needs to be split up
//
// RETURNS		method returns no data.
//===============================================================================

void WordList::MakeDB (char *theLine)
{
	char           *myWord;
	int             SourcePosition = 0;
	int             DestPosition = 0;
	int				wsz;
	char		   *tmp;

	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		Clear ();
		First = NULL;
		Last = NULL;

		while (theLine[SourcePosition] != 0)
		{
			if (tmp = strchr (theLine+SourcePosition, '|'))
			{
				wsz = tmp - (theLine+SourcePosition);
				while ((tmp) && (*tmp-1 == '\\'))
				{
					tmp = strchr (tmp, '|');
					if (tmp)
					{
						wsz = tmp - (theLine+SourcePosition);
					}
					else wsz = strlen (theLine);
				}
			}
			else wsz = strlen (theLine);
			
			myWord = new char[wsz+4];
			
			while ((theLine[SourcePosition] != 0) && (theLine[SourcePosition] != '|'))
			{
				if ((theLine[SourcePosition] == '\\') && (theLine[SourcePosition + 1] != 0))
					++SourcePosition;
				myWord[DestPosition++] = theLine[SourcePosition++];
			}

			if ((DestPosition > 2) && (myWord[DestPosition - 1] == ' '))
				--DestPosition;
			myWord[DestPosition] = 0;

			Add_ (myWord);
			
			DestPosition = 0;
			if (theLine[SourcePosition] != 0)
				++SourcePosition;
		}
	}
}

//===============================================================================
// METHOD		MakeGeneric
// SYNOPSIS		Turns a string into a wordlist. The seperation character to be
//				used is passed as an argument.
//
// ARGUMENTS	theLine		(char *)	String that needs to be split up
//				Seperator	(char)		Seperation character
//
// RETURNS		method returns no data.
//===============================================================================

void WordList::MakeGeneric (char *theLine, char Seperator)
{
	char           *myWord;
	int             SourcePosition = 0;
	int             DestPosition = 0;
	int				i;
	int				wsz;
	char		   *tmp;

	if ((theLine) && (Integrity[0]==13)&&(Integrity[1]==37))
	{
		Clear ();
		First = NULL;
		Last = NULL;

		while (theLine[SourcePosition] != 0)
		{
			if (tmp = strchr (theLine+SourcePosition, Seperator))
			{
				wsz = tmp - (theLine+SourcePosition);
				while ((tmp) && (*tmp-1 == '\\'))
				{
					tmp = strchr (tmp, Seperator);
					if (tmp)
					{
						wsz = tmp - (theLine+SourcePosition);
					}
					else wsz = strlen (theLine);
				}
			}
			else wsz = strlen (theLine);
			
			myWord = new char[wsz+4];
			myWord[0] = 0;
			while ((theLine[SourcePosition] != 0) && (theLine[SourcePosition] != Seperator))
			{
				if ((theLine[SourcePosition] == '\\') && (theLine[SourcePosition + 1] != 0))
					++SourcePosition;
				
				if (DestPosition < wsz)
					myWord[DestPosition++] = theLine[SourcePosition++];
				else ++SourcePosition;
			}

			if ((DestPosition > 2) && (myWord[DestPosition - 1] < 32))
				--DestPosition;
			myWord[DestPosition] = 0;

			Add_ (myWord);
			DestPosition = 0;
			if (theLine[SourcePosition] != 0)
				++SourcePosition;
		}
	}
}

//===============================================================================
// METHOD		Find
// SYNOPSIS		Returns the index of the first item in the list that matches
//				a string.
//
// ARGUMENTS	theMatch		(char *)	String to match
//
// RETURNS		index value. -2 for error
//===============================================================================

int WordList::Find (char *theMatch)
{
	Word           *myWord;

	int             result = -2;
	int             i = 0;

	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		if (theMatch)
		{
			myWord = First;

			while ((myWord) && (result < 0))
			{
				if (myWord->Contents)
				{
					if (strcasecmp (theMatch, myWord->Contents) == 0)
						result = i;
				}
				++i;
				myWord = myWord->Next;
			}
			return result;
		}
		else return -2;
	}
}

//===============================================================================
// METHOD		Add
// SYNOPSIS		Adds an item to the wordlist
//
// ARGUMENTS	theWord		(char *)	String to be added
//
// RETURNS		method returns no data.
//===============================================================================

void WordList::Add (char *theWord)
{
	int             len;

	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		if (theWord)
		{
			Word           *myWord = new Word;

			if (!myWord) printf ("out of memory!!\n");

			len = strlen (theWord);

			myWord->Contents = new char[len +2];

			if (strlen (theWord) == 0)
				myWord->Contents[0] = 0;
			else
				strncpy (myWord->Contents, theWord, strlen (theWord) + 1);

			++Count;

			if (First)
			{
				myWord->Prev = Last;
				myWord->Next = NULL;
				Last->Next = myWord;
				Last = myWord;
			}
			else
			{
				First = myWord;
				Last = myWord;
				myWord->Prev = NULL;
				myWord->Next = NULL;
			}
		}
	}
}

//===============================================================================
// METHOD		Add
// SYNOPSIS		Adds an item to the wordlist
//
// ARGUMENTS	theWord		(char *)	String to be added (will not be copied)
//
// RETURNS		method returns no data.
//===============================================================================

void WordList::Add_ (char *theWord)
{
	int             len;

	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		if (theWord)
		{
			Word           *myWord = new Word;

			myWord->Contents = theWord;

			++Count;

			if (First)
			{
				myWord->Prev = Last;
				myWord->Next = NULL;
				Last->Next = myWord;
				Last = myWord;
			}
			else
			{
				First = myWord;
				Last = myWord;
				myWord->Prev = NULL;
				myWord->Next = NULL;
			}
		}
	}
}

//===============================================================================
// METHOD		Trim
// SYNOPSIS		Trims a wordlist only to contain a maximum number of items
//
// ARGUMENTS	theNumber	(int)		Number of items the wordlist may
//										contain. If negative, cut from the end.
//
// RETURNS		Pointer to requested item. Pointer to empty string if item
//				was not found.
//===============================================================================

void WordList::Trim (int theNumber)
{
	int _theNumber;
	
	if (theNumber<0)
	{
		_theNumber = -theNumber;
		while (Count > _theNumber)
		{
			Delete (Count-1);
		}
	}
	else
	{
		_theNumber = theNumber;
		while (Count > _theNumber)
		{
			Delete (0);
		}
	}
}

//===============================================================================
// METHOD		Get
// SYNOPSIS		Returns a numbered item
//
// ARGUMENTS	theNumber	(int)		Index value
//
// RETURNS		Pointer to requested item. Pointer to empty string if item
//				was not found.
//===============================================================================

char *WordList::Get (int theNumber)
{
	int             i = 0;
	Word           *myWord;

	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		myWord = First;

		if ((Cache) && (theNumber==CachePos))
		{
			if (Cache->Contents) return Cache->Contents;
			else if (Cache->WContents)
			{
				return (Cache->WContents->Get (0));
			}
			else if (Cache->IntSet)
			{
				Cache->Contents = new char[32];
				sprintf (Cache->Contents, "%i", Cache->IContents);
				return (Cache->Contents);
			}
		}
		if (theNumber==(CachePos+1))
		{
			if (Cache)
			{
				Cache = Cache->Next;
				++CachePos;
				if (Cache)
				{
					if (Cache->Contents) return (Cache->Contents);
					else if (Cache->WContents) return (Cache->WContents->Get(0));
					else if (Cache->IntSet)
					{
						Cache->Contents = new char[32];
						sprintf (Cache->Contents, "%i", Cache->IContents);
						return (Cache->Contents);
					}
				}
				else
				{
					return (char *) "";
					CachePos = -3;
				}
			}
		}

		while ((myWord) && (i < theNumber))
		{
			myWord = myWord->Next;
			++i;
		}

		if (myWord)
		{
			Cache = myWord;
			CachePos = theNumber;
			if (myWord->Contents) return myWord->Contents;
			else if (myWord->WContents) return myWord->WContents->Get(0);
			else if (myWord->IntSet)
			{
				myWord->Contents = new char[32];
				sprintf (myWord->Contents, "%i", myWord->IContents);
				return (myWord->Contents);
			}
		}
		else
			return (char *) "";
	}
	return (char *) "";
}

//===============================================================================
// METHOD		Delete
// SYNOPSIS		Removes a numbered item from the list
//
// ARGUMENTS	theNumber	(int)		Index value
//
// RETURNS		Method returns no data.
//===============================================================================

void WordList::Delete (int theNumber)
{
	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		int             i = 0;
		Word           *myWord;

		myWord = First;

		while ((myWord) && (i < theNumber))
		{
			myWord = myWord->Next;
			++i;
		}

		if (myWord)
		{
			--Count;
			Cache = NULL;
			CachePos = -2;
			if (myWord->Next)
			{
				if (myWord->Prev)
				{
					myWord->Prev->Next = myWord->Next;
					myWord->Next->Prev = myWord->Prev;
				}
				else
				{
					myWord->Next->Prev = NULL;
					First = myWord->Next;
				}
			}
			else
			{
				if (myWord->Prev)
				{
					myWord->Prev->Next = NULL;
					Last = myWord->Prev;
				}
				else
				{
					First = NULL;
					Last = NULL;
				}
			}
			delete (myWord);
		}
	}
}

//===============================================================================
// METHOD		Set
// SYNOPSIS		Replaces the contents of a numbered item with something else.
//
// ARGUMENTS	theNumber	(int)		Index value
//				theContents	(char *)	New string
//
// RETURNS		Method returns no data.
//===============================================================================

void WordList::Set (int theNumber, char *theContents)
{
	int             i = 0;
	Word           *myWord;


	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		if (theNumber >= Count)
		{
			while (Count <= theNumber)
				Add (".");
		}

		myWord = First;

		while ((myWord) && (i < theNumber))
		{
			myWord = myWord->Next;
			++i;
		}

		if (myWord)
		{
			if (myWord->WContents)
			{
				delete (myWord->WContents);
				myWord->WContents = NULL;
			}
			if (myWord->Contents != NULL) delete[] (myWord->Contents);
			myWord->Contents = new char[strlen (theContents) + 3];

			strcpy (myWord->Contents, theContents);
		}
	}
}

//===============================================================================
// METHOD		SetInt
// SYNOPSIS		Replaces the contents of a numbered item with an integer.
//
// ARGUMENTS	theNumber	(int)		Index value
//				theContents	(int)		New integer
//
// RETURNS		Method returns no data.
//===============================================================================

void WordList::SetInt (int theNumber, int I)
{
	int             i = 0;
	Word           *myWord;


	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		myWord = First;

		if (theNumber >= Count)
		{
			while (Count <= theNumber)
				Add (".");
		}

		while ((myWord) && (i < theNumber))
		{
			myWord = myWord->Next;
			++i;
		}

		if (myWord)
		{
			if (myWord->WContents)
			{
				delete (myWord->WContents);
				myWord->WContents = NULL;
			}
			if (myWord->Contents) delete[] (myWord->Contents);
			myWord->Contents = NULL;
			myWord->IntSet = 1;
			myWord->IContents = I;
		}
	}
}

//===============================================================================
// METHOD		Gather
// SYNOPSIS		Turns the wordlist back into a string.
//
// ARGUMENTS	Start		(int)		Index value of first item
//				Buffer		(char *)	Storage for the string
//				maxlen		(int)		Size of the buffer
//
// RETURNS		Returns pointer to Buffer.
//===============================================================================

char *WordList::Gather (int Start, char *Buffer, int maxlen)
{
	int             i = Start;

	if (Buffer)
	{

		Buffer[0] = 0;

		if ((Integrity[0]==13)&&(Integrity[1]==37))
		{
			while (i < Count)
			{
				strncat (Buffer, Get (i++), maxlen);
				if (i < Count)
					strncat (Buffer, " ", maxlen);
			}

			Buffer[maxlen] = 0;
		}
		return Buffer;
	}
	else
	{
		int totlen;
	}
}

//===============================================================================
// METHOD		Clear
// SYNOPSIS		Deletes all items in the list.
//
// RETURNS		Method returns no data.
//===============================================================================

void WordList::Clear (void)
{
	Word           *W = First;
	Word           *WW;

	Cache = NULL;
	CachePos = -3;

	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		while (W)
		{
			WW = W->Next;
			delete (W);
			W = WW;
		}

		First = NULL;
		Last = NULL;

		Count = 0;
	}
}

void WordList::AddInt (int I)
{
	Word *Wo = new Word;
	Wo->IContents = I;
	Wo->IntSet = 1;
	
	if (Last)
	{
		Wo->Prev = Last;
		Wo->Next = NULL;
		Last->Next = Wo;
		Last = Wo;
	}
	else
	{
		Wo->Prev = NULL;
		Wo->Next = NULL;
		First = Last = Wo;
	}
	
	++Count;
}

void WordList::AddWordList (WordList *W)
{
	WordList *WW = new WordList (W);
	AddWordList_ (WW);
}

void WordList::AddWordList_ (WordList *W)
{
	Word *Wo = new Word;
	Wo->WContents = W;
	
	if (Last)
	{
		Wo->Prev = Last;
		Wo->Next = NULL;
		Last->Next = Wo;
		Last = Wo;
	}
	else
	{
		Wo->Prev = NULL;
		Wo->Next = NULL;
		First = Last = Wo;
	}
	
	++Count;
}

int WordList::GetInt (int theNumber)
{
	int		  i=0;
	Word		*myWord;
	
	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		myWord = First;
		
		if ((Cache) && (theNumber == CachePos))
		{
			if (Cache->IntSet) return Cache->IContents;
			if (Cache->Contents)
			{
				Cache->IntSet = 1;
				Cache->IContents = atoi (Cache->Contents);
				return Cache->IContents;
			}
			if (Cache->WContents)
			{
				Cache->IntSet = 1;
				Cache->IContents = atoi (Cache->WContents->Get(0));
				return Cache->IContents;
			}
		}
		if (theNumber==(CachePos+1))
		{
			if (Cache)
			{
				Cache = Cache->Next;
				++CachePos;
				if (Cache)
				{
					if (Cache->IntSet) return Cache->IContents;
					if (Cache->Contents)
					{
						Cache->IntSet = 1;
						Cache->IContents = atoi (Cache->Contents);
						return Cache->IContents;
					}
					if (Cache->WContents)
					{
						Cache->IntSet = 1;
						Cache->IContents = atoi (Cache->WContents->Get(0));
						return Cache->IContents;
					}
				}
				else
				{
					return NULL;
					CachePos = -3;
				}
			}
		}

		while ((myWord) && (i < theNumber))
		{
			myWord = myWord->Next;
			++i;
		}

		if (myWord)
		{
			Cache = myWord;
			CachePos = theNumber;
			if (myWord->IntSet) return myWord->IContents;
			if (myWord->Contents)
			{
				myWord->IntSet = 1;
				myWord->IContents = atoi (myWord->Contents);
				return myWord->IContents;
			}
			if (myWord->WContents)
			{
				myWord->IntSet = 1;
				myWord->IContents = atoi (myWord->WContents->Get(0));
				return myWord->IContents;
			}
		}
		else
			return 0;
	}
	return 0;
}

WordList *WordList::GetWordList (int theNumber)
{
	int             i = 0;
	Word           *myWord;

	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		myWord = First;

		if ((Cache) && (theNumber==CachePos))
		{
			if (Cache->WContents) return Cache->WContents;
			else if (Cache->Contents)
			{
				Cache->WContents = new WordList;
				Cache->WContents->MakeDB (Cache->Contents);
				return (Cache->WContents);
			}
			else if (Cache->IntSet)
			{
				Cache->Contents = new char[32];
				sprintf (Cache->Contents, "%i", Cache->IContents);
				Cache->WContents = new WordList;
				Cache->WContents->MakeDB (Cache->Contents);
				return (Cache->WContents);
			}
		}
		if (theNumber==(CachePos+1))
		{
			if (Cache)
			{
				Cache = Cache->Next;
				++CachePos;
				if (Cache)
				{
					if (Cache->WContents) return (Cache->WContents);
					else if (Cache->Contents)
					{
						Cache->WContents = new WordList;
						Cache->WContents->MakeDB (Cache->Contents);
						return (Cache->WContents);
					}
					else if (Cache->IntSet)
					{
						Cache->Contents = new char[32];
						sprintf (Cache->Contents, "%i", Cache->IContents);
						Cache->WContents = new WordList;
						Cache->WContents->MakeDB (Cache->Contents);
						return (Cache->WContents);
					}
				}
				else
				{
					return NULL;
					CachePos = -3;
				}
			}
		}

		while ((myWord) && (i < theNumber))
		{
			myWord = myWord->Next;
			++i;
		}

		if (myWord)
		{
			Cache = myWord;
			CachePos = theNumber;
			if (myWord->WContents) return myWord->WContents;
			else if (myWord->Contents)
			{
				myWord->WContents = new WordList;
				myWord->WContents->MakeDB (myWord->Contents);
				return myWord->WContents;
			}
			else if (myWord->IntSet)
			{
				myWord->Contents = new char[32];
				sprintf (myWord->Contents, "%i", myWord->IContents);
				myWord->WContents = new WordList;
				myWord->WContents->MakeDB (Cache->Contents);
				return (myWord->WContents);
			}
		}
		else
			return NULL;
	}
	return NULL;
}

int WordList::IsWordList (int theNumber)
{
	int             i = 0;
	Word           *myWord;

	if ((Integrity[0]==13)&&(Integrity[1]==37))
	{
		myWord = First;

		if ((Cache) && (theNumber==CachePos))
		{
			if (Cache->WContents) return 1;
			return 0;
		}
		if (theNumber==(CachePos+1))
		{
			if (Cache)
			{
				Cache = Cache->Next;
				++CachePos;
				if (Cache)
				{
					if (Cache->WContents) return 1;
					return 0;
				}
				else
				{
					return 0;
					CachePos = -3;
				}
			}
		}

		while ((myWord) && (i < theNumber))
		{
			myWord = myWord->Next;
			++i;
		}

		if (myWord)
		{
			Cache = myWord;
			CachePos = theNumber;
			if (myWord->WContents) return 1;
			return 0;
		}
		else
			return 0;
	}
	return 0;
}
